package ctree;

import main.CValues;
import main.Values;

import java.io.*;
import java.nio.ByteBuffer;
import java.util.*;

/**
 * A disk-based, paginated and buffered index.
 * The index reads and writes CPage using a {@link LRUBuffer}.
 * <p>
 * <b>Usage:</b><br>
 * <ul>
 * <li>The file can be viewed as a consecutive list of disk pages, each of size
 * pageSize. See the figure below:<br>
 * <pre>
 * +------+-------+-------+-------+-------+-------+-------+
 * | head | page1 | page2 | page3 | page4 | page5 | page6 |
 * +------+-------+-------+-------+-------+-------+-------+
 * </pre>
 * <li>
 * The head page (having pageID=0) uses the first OVERHEAD=16 bytes to store
 * four integers:<br>
 *   <ul>
 *   <li><b>-1</b>
 *   <li><b>pagesize</b>
 *   <li><b>numOfPages</b>: number of non-empty pages
 *   <li><b>free</b>: pageID of the first empty page
 *   </ul>
 * The rest pageSize - OVERHEAD bytes stores the user-specified head information.
 * For instance, a tree index may store the pageID for the root page here. To manage
 * the head information, a derived class needs to implement the three abstract functions
 *   <ul>
 *   <li><code>initIndexHead</code>: called when the index is built;
 *   <li><code>readIndexHead</code>: called when the index is opened;
 *   <li><code>writeIndexHead</code>: called right before the index is closed.
 *   </ul>
 * <li>
 * Besides the constructor and the <code>close()</code> function, typical usages
 * of the <code>CIndex</code> are:
 *   <ul>
 *   <li><code>allocate</code>: to get a new pageID;
 *   <li><code>freePage</code>: to declare a pageID to be empty;
 *   <li>The public functions in the buffer, e.g. <code>readPage</code>,
 *       <code>writePage</code>.
 *   </ul>
 * <li> One implementation detail is that
 * all the empty pages in the file are linked together in
 * a linked list. An empty page contains two integers: 0 and pageID for the
 * next empty page. If it is the last empty page in the file,
 * the next pointer is set to -1. A class called DBEmptyPage is defined for this purpose.
 * But all these are hidden from the user.
 * </ul>
 *
 * @see CPage
 * @see LRUBuffer
 * @author Mikalai Tsytsarau
 */

public class CIndex extends LRUBuffer implements Iterable<CPage> {
	public static final boolean CREATE = true;
	public static final boolean OPEN = false;

    //pageID of the last page
	protected int rootPageID;
	//the last page
	public CPage rootPage;
    //root page height
    public int height; //maximum height of the root node
    public int minInterval; //leaf node interval length
    public int numIntervals; //data frequency at leaf nodes
    //The buffer manager.
    private ByteBuffer buffer;
	//The number of non-empty pages (excluding the head page).
	public int numPages = 0;
	// The first empty page ID. It is -1 if there is no empty page in the file.
	protected int firstEmpty = -1;

    public boolean sorted;
    public boolean extended;
    public boolean compressed;
    public boolean multilevel;

    private int iteratorHeight = 1;
    private int iteratorBegin;
    private int iteratorEnd;

    /**
	 * Creates or Opens an index file.
	 *
	 * @param file  the filename
	 * @param isCreate    whether create (or open)
     * @param bufferSize buffer size in pages
	 * @param sorted if topics are sorted
     * @param extended if demographics
     * @param compressed if normalized values (1 byte)
     * @param minInterval
     * @param begin_time begin time of the root page
     * @param end_time end time of the root page
     * @throws IOException if read/write error
	 */
    public CIndex(File file, boolean isCreate, int bufferSize, int pageSize, boolean sorted, boolean extended, boolean compressed, boolean multilevel, int minInterval, int numIntervals, int begin_time, int end_time) throws IOException {

        super(file, bufferSize, pageSize);
        buffer = ByteBuffer.allocateDirect(pageSize);

        this.sorted = sorted;
        this.extended = extended;
        this.compressed = compressed;
        this.multilevel = multilevel;
        this.minInterval = minInterval;
        this.numIntervals = numIntervals;
        this.height = intervalGranularity(begin_time, end_time);

		if (isCreate) {
            rootPage = (CPage) newPage(DBPage.TYPE_CTREE);
            rootPage.init(-1, height, begin_time, end_time);
            rootPageID = rootPage.pageId;
            writePage(rootPage);
		} else {
			DBByteArrayPage head = (DBByteArrayPage) readPage(0);
			ByteBuffer headPage = ByteBuffer.wrap(head.content);
			readIndexHead(headPage);
            if (rootPageID == 0) { rootPageID = 1; System.err.println("error reading index head");}
            rootPage = (CPage) readPage(rootPageID);//1
            //System.out.println(rootPage.topics.size());
		}

        this.iteratorHeight = rootPage.height;
        this.iteratorBegin = rootPage.begin;
        this.iteratorEnd = rootPage.end;
    }

    public void compress() throws IOException {
        CIndex index = new CIndex(new File(fileName.replaceAll(".dat", "_compressed.dat")), true, bufferSize, pageSize, false, false, true, false, minInterval, numIntervals, rootPage.begin, rootPage.end);
        for (int i = 1; i <= numPages; i++) {
            DBPage page = readPage(i);
            if (page != null && (page.pageType == DBPage.TYPE_CTREE || page.pageType == DBPage.TYPE_CTREE_EXTENT)) {
                CPageExtent p = (CPageExtent) page;
                if (p.height == 0) {
                    List<CValues> vals = p.getPageValues();
                    for (CValues v : vals) {
                        index.insert(v.id, v.time(), v, 0);
                    }
                    //System.out.print(i + ":" + page.pageType + " ");
                }
            }
        }
        index.close();
    }

    public CIndex(File file) throws IOException {
         this(file, true, 1024, 4096, false, false, false, false, CModel.DAY, 1, 0, 10);
    }

	/**
	 * Assigns a new pageID.
	 *
	 * @return the new page ID.
	 */
	protected int allocate() {
		int newID = ++numPages; //TODO: remove this, use last ID
		if (firstEmpty >= 0) {
            newID = firstEmpty;
            // get new firstEmpty
            DBPage emptyPage = readPage(firstEmpty);
            if (emptyPage != null && emptyPage.pageType == 0) {
                firstEmpty = ((DBEmptyPage)emptyPage).nextEmptyPage;
            } else {
                firstEmpty = -1;
                System.err.println("Problem allocating next empty page id");
            }
        }
		return newID;
	}

	/**
	 * Frees an empty page.
	 *
	 * @param pageID the to-be-freed page ID
	 * @throws IOException if write error
	 */
	public void freePage(int pageID) throws IOException {
        DBEmptyPage emptyPage = new DBEmptyPage(this, pageSize, pageID, firstEmpty);
		writePage(emptyPage);
		firstEmpty = pageID;
		numPages--;
	}

    /**
	 * Writes a DBPage to buffer.
	 *
	 * @param page the DBPage object
	 */
	public void writePage(DBPage page) {
		bufferWriteIO++;
		CBufferElement stored = find(page.pageId);
		if (stored == null) {
			stored = new CBufferElement();
			stored.pageID = page.pageId;
            try {
                add(stored);
            } catch (IOException e) {
                System.err.println("Unable to store in a cache page " + stored.pageID + ":");
                e.printStackTrace();
            }
		}
        stored.object = page;
        stored.parsed = true;
        stored.dirty = true;
        stored.nodeType = page.pageType;
	}

    protected void savePage(CBufferElement page) throws IOException {
        if (page.dirty) {
            buffer.position(0);
            ((DBPage)page.object).write(buffer);
            // fill with zeroes
            while (buffer.hasRemaining()) {
                buffer.put((byte)0);
            }
            save(buffer, page.pageID);
            page.dirty = false;
        }
    }

    /**
	 * Saves the header page to disk and flushes the buffer.
	 * It is important to call <code>close()</code> before exiting.
	 * Otherwise the file may be corrupted.
	 *
	 * @throws IOException if write error
	 */
	public void close() throws IOException {
        buffer.position(0);
        writeIndexHead(buffer);
        DBByteArrayPage page = new DBByteArrayPage(this, pageSize, 0, buffer);
        writePage(page);
        super.close();
	}

    /**
     * Reads a page from buffer and parse it if not parsed already.
	 *
	 * @param pageType specifies page type
     * @return  the parsed page
	 */
	public DBPage newPage(int pageType) {
        int pageID = allocate();
        DBPage page = null;
        switch (pageType) {
            case DBPage.TYPE_ARRAY: page = new DBByteArrayPage(this, pageSize, pageID);
                break;
            case DBPage.TYPE_EMPTY: page = new DBEmptyPage(this, pageSize, pageID);
                break;
            case DBPage.TYPE_CTREE:
                page = new CPage(this, pageSize, pageID);
                break;
            case DBPage.TYPE_CTREE_EXTENT:
                page = new CPageExtent(this, pageSize, pageID);
                break;
        }
        writePage(page);
		return page;
	}

    /**
	 * Returns a buffered page. If it is in buffer, returns it. Otherwise,
	 * loads it from disk and then returns it. Increases <code>bufferReadIO</code>.
	 *
	 * @param pageID page ID
	 * @return the buffered page, null if read error
	 */
	public DBPage readPage(int pageID) {
        if (pageID < 0) return null;
        //System.out.print(pageID + " ");
        bufferReadIO++;
		CBufferElement stored = find(pageID);
		if (stored != null) return (DBPage) stored.object;

        buffer.position(0);
        try {
            load(buffer, pageID);
        } catch (IOException e) {
            System.err.println("Unable to load page " + pageID + ":");
            e.printStackTrace();
            return null;
        }
        DBPage page = null;
        stored = new CBufferElement();
        stored.parsed = true;
        stored.dirty = false;
        stored.nodeType = buffer.get();
        stored.pageID = pageID;

        switch (stored.nodeType) {
            case DBPage.TYPE_ARRAY: page = new DBByteArrayPage(this, pageSize, pageID);
                break;
            case DBPage.TYPE_EMPTY: page = new DBEmptyPage(this, pageSize, pageID);
                break;
            case DBPage.TYPE_CTREE:
                page = new CPage(this, pageSize, pageID);
                break;
            case DBPage.TYPE_CTREE_EXTENT:
                page = new CPageExtent(this, pageSize, pageID);
            default:
                return page;
        }
        buffer.position(0);
        try {
            page.read(buffer);
        } catch (Exception e) {
            System.err.println("Unable to parse a page " + stored.pageID + ": type " + stored.nodeType);
            e.printStackTrace();
            return null;
        }
        stored.object = page;
        try {
            add(stored);
        } catch (IOException e) {
            System.err.println("Unable to store in a cache page " + stored.pageID + ":");
            e.printStackTrace();
        }
//      if (page.height >= 2) {
//           buffer.pin(file, page.pageId);
//      }
		return page;
	}

	/**
	 * Reads index head information from a byte array.
	 * Called when opening a CIndex.
	 * @param bf the byte array of length pageSize
	 */
	protected void readIndexHead(ByteBuffer bf) {
		byte pageType = bf.get();
        if (pageType != -1) {
			System.out.println("Illegal file opened!"); System.exit(0);
		}
		pageSize = bf.getInt();
        numPages = bf.getInt();
        firstEmpty = bf.getInt();
        rootPageID = bf.getInt();
        sorted = (bf.get() > 0);
        extended = (bf.get() > 0);
        //compressed = (bf.get() > 0);
	}

	/**
	 * Writes index head information to a byte array.
	 * Called before closing a DBIndex.
	 * @param bf the byte array of length pageSize
	 */
	protected void writeIndexHead(ByteBuffer bf) {
		bf.put((byte) -1);
		bf.putInt(pageSize);
		bf.putInt(numPages);
		bf.putInt(firstEmpty);
		bf.putInt(rootPageID);
        bf.put((byte) (sorted ? 1 : 0));
        bf.put((byte) (extended ? 1 : 0));
        //bf.put((byte) (compressed ? 1 : 0));
	}

    public void insert(int topicId, int timestamp, float s) {
        insert(topicId, timestamp, s, 0);
    }

    public void insert(int topicId, int timestamp, float s, int demographics_id) {
        insert(topicId, timestamp, new Values(1, s, s * s), demographics_id);
    }

    public void insert(int topicId, int timestamp, Values values, int demographics_id) {
        if (timestamp <= rootPage.end & timestamp > rootPage.begin) {
            rootPage.insert(topicId, timestamp, values, demographics_id);
            //allTopics.remove(topic);
            //allTopics.add(topic);
            //numTopics = allTopics.size();
            //if (topicId > maxTopicId) maxTopicId = topicId;
        }
    }

    public void sort() throws IOException {
        if (!sorted) {
            rootPage.sort();
            sorted = true;
        }
    }

    public CPage getFirstPage(int height, int begin, int end) {
        //looks for the first non-empty page containing or after begin
        //descends for children, until height is reached
        int timestamp = begin;
        CPage page = rootPage;
        while (page != null && page.height > height) {
            int child_id = page.findChild(timestamp);
            if (child_id < 0) {
                child_id = 0; //reset to search from beginning of new page
            }
            // move forward to find first non-empty page
            page = (CPage) readPage(page.findNeighbour(child_id - 1, 1));
            if (page != null && page.begin >= end) {
                page = null;
            }
        }
        return page;
    }

    public CPage getLastPage(int height, int begin, int end)  {
        //looks for the last non-empty page containing or before end
        //descends for children, until height is reached
        int timestamp = end;
        CPage page = rootPage;
        while (page != null && page.height > height) {
            int child_id = page.findChild(timestamp);
            if (child_id >= page.numChildren) {
                child_id = page.numChildren - 1; //reset to search from end of new page
            }
            page = (CPage) readPage(page.findNeighbour(child_id + 1, -1));
            if (page != null && page.end <= begin) {
                page = null;
            }
        }
        return page;
    }

    public void setInterval(int height, int begin, int end) {
        iteratorHeight = height;
        iteratorBegin = begin;
        iteratorEnd = end;
    }

    @Override
    public Iterator<CPage> iterator() {
        return new CPageIterator(this, iteratorHeight, iteratorBegin, iteratorEnd);
    }

    //TODO: fix this three below
    public int intervalGranularity(int begin, int end) {
        if (!multilevel || minInterval == CModel.YEAR) {
            return (int) Math.ceil(Math.log10(((double) end - (double) begin) / minInterval));
        } else {
            return 1;
        }
    }

    public int getInterval(int height) {
        return (int) Math.pow(10, height) * minInterval * (height == 0 ? numIntervals : 1);
    }

    public byte getNumChildren(int height) {
        if (height == 0) return 0;
        if (!multilevel || minInterval == CModel.YEAR) {
            return (byte) 10;
        } else {
            return (byte) 10;
        }
    }

    public class CPageIterator implements Iterator<CPage> {
        private CIndex index;
        private CPage page;
        private int end;

        public CPageIterator (CIndex index, int height, int begin, int end) {
            this.index = index;
            this.page = index.getFirstPage(height, begin, end);
            this.end = end;
        }

        @Override
        public boolean hasNext() {
            return (page != null && page.end < end);
        }

        @Override
        public CPage next() {
            CPage next = page;
            page = (CPage) index.readPage(page.next);
            return next;
        }

        @Override
        public void remove() {
            //we can not remove pages just yet
        }
    }

    public static void main(String[] args) throws Exception {
        int begin = (int)(CModel.formatDay.parse("01.01.1912").getTime() / 1000L);
        int end = (int)(CModel.formatDay.parse("01.01.2012").getTime() / 1000L);
        CIndex i = new CIndex(null, true, 1024, 4096, false, false, true, false, CModel.YEAR, 1, begin, end);
        System.out.println(i.rootPage);
        CPage p1 = (CPage) i.newPage(DBPage.TYPE_CTREE); p1.init(-1, 0, 0, 0);
        System.out.println(p1.pageCapacity + " topics" );
        CPageExtent p2 = (CPageExtent) i.newPage(DBPage.TYPE_CTREE_EXTENT); p2.init(0, 0, 0);
        System.out.println(p2.pageCapacity + " topics" );
    }
}