package ctree;

/**
 * A class for managing {@link ctree.CIndex} and retrieving contradictions.
 * It can read, write and update sentiment values from a CTree via {@link ctree.Model} interface.
 *
 * @author Mikalai Tsytsarau;
 */

import main.*;
import ui.PropertiesManager;
import java.io.*;
import java.text.SimpleDateFormat;
import java.util.*;

public class CModel implements Model {
    public static final int DAY = 86400;
    public static final int WEEK = 86400 * 7;
    public static final int MONTH = (int)(86400 * 30.5);
    public static final int YEAR = (int)(86400 * 365.25);


    public static final int INDEX_VER_0 = 0;
    public static final int INDEX_VER_1 = 1;

    public CIndex index = null;
    public static String indexPath = "./index/";
    public static String indexName = "index.dat";
    public static int indexVersion = INDEX_VER_1;
    public boolean indexSorted = false;
    public boolean indexExtended = false;
    public boolean indexMultilevel = false;
    public boolean indexCompressed = false;
    public int bufferSize = 1024;
    public int pageSize = 4096;

    public boolean adaptive = false;
    public float threshold = 1f;
    public float smoothing = 1f;
    public int indexBegin;
    public int indexEnd;
    public int intervalBegin = 0;
    public int intervalEnd = Integer.MAX_VALUE;
    public int window = 10;
    public int aggregate = 1;
    public int granularity = 1;
    public int maxGranularity = 1;
    public int minInterval = 86400;
    public int topicId = -1;
    public int sliding_range = 2; // needs to be even

    public CPage timeNode;

    public static SimpleDateFormat formatHour = new SimpleDateFormat("dd.MM HH:mm");
    public static SimpleDateFormat formatDay = new SimpleDateFormat("dd.MM.yyyy");
    public static SimpleDateFormat formatMonth = new SimpleDateFormat("MM.yyyy");
    public static SimpleDateFormat formatYear = new SimpleDateFormat("yyyy");

    public HashMap<Integer,TimeSeries> buckets;
    public int contrType = CValues.TYPE_CONTR_HIGH;
    public PropertiesManager properties;

    public static boolean statsAgainstAvg = false;
    private String minIntervalType = "day";
    public int numIntervals = 1;

    public CModel () {
        properties = new PropertiesManager();
        loadConfig();
        readConfig();
    }

    public void loadConfig() {
        properties.load();
    }

    public void readConfig() {
        System.out.println("Read Config");
        indexPath = properties.getProperty(PropertiesManager.INDEX, "index_path");
        indexName = properties.getProperty(PropertiesManager.INDEX, "index_name");
        bufferSize = Integer.parseInt(properties.getProperty(PropertiesManager.INDEX, "bufferSize"));
        pageSize = Integer.parseInt(properties.getProperty(PropertiesManager.INDEX, "pageSize"));
        try {
            indexBegin = (int) (formatDay.parse(properties.getProperty(PropertiesManager.INDEX, "begin")).getTime() / 1000);
            indexEnd = (int) (formatDay.parse(properties.getProperty(PropertiesManager.INDEX, "end")).getTime() / 1000);
        } catch (Exception e) {
            System.err.println("Index time interval is not specified in config!");
        }
        try {
            maxGranularity = Integer.parseInt(properties.getProperty(PropertiesManager.INDEX, "maxGranularity"));
            numIntervals = Integer.parseInt(properties.getProperty(PropertiesManager.INDEX, "numIntervals"));
            minIntervalType = properties.getProperty(PropertiesManager.INDEX, "minInterval");
            if (minIntervalType.equalsIgnoreCase("day")) {
                minInterval = DAY;
            } else if (minIntervalType.equalsIgnoreCase("week")) {
                minInterval = WEEK;
            } else if (minIntervalType.equalsIgnoreCase("month")) {
                minInterval = MONTH;
            } else if (minIntervalType.equalsIgnoreCase("year")) {
                minInterval = YEAR;
            }
        } catch (Exception e) {
            System.err.println("Min interval is not specified in config!");
        }
        indexVersion = Integer.parseInt(properties.getProperty(PropertiesManager.INDEX, "version", "1"));
        indexSorted = Boolean.parseBoolean(properties.getProperty(PropertiesManager.INDEX, "sorted")); //needed only for the creation of new index
        indexCompressed = Boolean.parseBoolean(properties.getProperty(PropertiesManager.INDEX, "compressed"));
        indexMultilevel = Boolean.parseBoolean(properties.getProperty(PropertiesManager.INDEX, "multilevel"));
        indexExtended = Boolean.parseBoolean(properties.getProperty(PropertiesManager.INDEX, "dIndexEnabled"));
    }

    public void updateConfig() {
        System.out.println("Update Config");
        properties.putProperty(PropertiesManager.INDEX, "version", indexVersion + "");
        properties.putProperty(PropertiesManager.INDEX, "index_path", indexPath);
        properties.putProperty(PropertiesManager.INDEX, "index_name", indexName);
        properties.putProperty(PropertiesManager.INDEX, "sorted", indexSorted + "");
        properties.putProperty(PropertiesManager.INDEX, "compressed", indexCompressed + "");
        properties.putProperty(PropertiesManager.INDEX, "multilevel", indexMultilevel + "");
        properties.putProperty(PropertiesManager.INDEX, "maxGranularity", maxGranularity + "");
        properties.putProperty(PropertiesManager.INDEX, "minInterval", minIntervalType);
        properties.putProperty(PropertiesManager.INDEX, "numIntervals", numIntervals + "");


    }

    public void storeConfig() throws IOException {
        System.out.println("Store Config");
        properties.store();
    }

    public void InitializeIndex(File file) throws IOException {
        if (file == null && !indexPath.equalsIgnoreCase("memory")) {
            file = new File(indexPath + indexName);
        } else {
            bufferSize *= 2048;
            System.out.println("Memory index: " + indexName);
        }
        boolean isCreate = true;
        if (file != null) {
            indexPath = "./" + (new File(".")).toURI().relativize(file.getParentFile().toURI()).getPath();
            indexName = file.getName();
            updateConfig();
           // file.delete();
            isCreate = !file.exists();
            if (isCreate) {
                System.out.println("Creating index: " + file.getName());
            } else {
                System.out.println("Opening index: " + file.getName());
                updateTopics();
            }
        }

        index = new CIndex(file, isCreate, bufferSize, pageSize, indexSorted, indexExtended, indexCompressed, indexMultilevel, minInterval, numIntervals, indexBegin, indexEnd);
        //index.rewrite();
    }


    public void FreeResources() {
        try {
            index.close();
            properties.store();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }


    public List<CValues> getContradictingTopics(int begin, int end, int height, float threshold, boolean isRecursive, int limit) {
        List<CValues> contradicting_topics = new ArrayList<CValues>();
        List<Integer> contradicting_ids = new ArrayList<Integer>();
        if (!isRecursive) {
            index.setInterval(height, begin, end);
            for (CPage page : index) {
                List<CValues> topics = page.getAllValues(limit);
                for (CValues topic : topics) {
                    if (topic.c > threshold && !contradicting_ids.contains(topic.id)) {
                        contradicting_ids.add(topic.id);
                        contradicting_topics.add(topic);
                    }
                }
                if (contradicting_topics.size() >= 1000) break;
            }
        } else {
            index.setInterval(height + 1, begin, end);
            for (CPage page : index) {
                List<CValues> parent_topics = page.getAllValues(limit);
                for (int i = 0; i < page.numChildren; i++) {
                    if (page.subpages.get(i) > 0 && page.timestamps.get(i) > begin && (page.timestamps.get(i) - page.subInt) < end) {
                        CPage child_page = (CPage) index.readPage(page.subpages.get(i));
                        List<CValues> child_topics = child_page.getAllValues(limit);
                        for (int j = 0; j < child_topics.size(); j++) {
                            CValues topic = child_topics.get(j);
                            CValues parent_topic = parent_topics.get(j);
                            if (!contradicting_ids.contains(topic.id) && topic.c > parent_topic.c * threshold) {
                                contradicting_ids.add(topic.id);
                                contradicting_topics.add(topic);
                            }
                        }
                    }
                }
                if (contradicting_topics.size() >= 1000) break;
            }
        }
        return contradicting_topics;
    }

    public List<CValues> getContradictingTopicsTopK(int begin, int end, int height, int k) {
        List<CValues> topk = new ArrayList<CValues>();
        index.setInterval(height, begin, end);
        for (CPage page : index) {
            List<CValues> topics = page.getTopics(indexSorted ? k : -1);
            for (CValues topic : topics) {
                //update top k
                int pos_topic = Integer.MAX_VALUE;
                boolean added_topic = false;
                boolean removed_topic = false;
                //find a place to insert
                for (int j = 0; j < topk.size(); j++) {
                    if (topic.id == topk.get(j).id) {
                        pos_topic = j;
                        if (added_topic) {
                            topk.remove(j);
                            removed_topic = true;
                            break;
                        }
                    }
                    //if topic has a higher ct - insert
                    if (!added_topic && topic.c > topk.get(j).c) {
                        //insert only if same topic with a higher ct not present
                        if (pos_topic >= j) {
                            topk.add(j, topic);
                            pos_topic = j;
                        }
                        added_topic = true;
                    }
                }
                if (!added_topic && pos_topic > topk.size() && topk.size() < k) {
                    topk.add(topic);
                }
                if (!removed_topic && topk.size() > k) {
                    topk.remove(topk.size() - 1);
                }
            }
        }
        return topk;
    }

    public List<CValues> getContradictionsTopK(int topicId, int begin, int end, int height, int k) {
        List<CValues> topk = new ArrayList<CValues>();
        float minValue = Float.NEGATIVE_INFINITY;
        index.setInterval(height, begin, end);
        for (CPage page : index) {
            boolean added = false;
            CValues v = page.getValuesTopK(topicId, minValue);
            if (v != null) {
                for (int j = 0; j < topk.size(); j++) {
                    if (v.c >= topk.get(j).c) {
                        topk.add(j, v);
                        added = true;
                        break;
                    }
                }
                if (topk.size() > k) {
                    topk.remove(topk.size() - 1);
                }
                if (!added && topk.size() < k) {
                    topk.add(v);
                }
                minValue = topk.get(topk.size() - 1).c;
            }
        }
        return topk;
    }

    public TimeSeries getContradictions(int topicId, int begin, int end, int height, float threshold, boolean isAdaptive) {
        HashMap<Integer,TimeSeries> series = new HashMap<Integer,TimeSeries>();
        series.put(0, new TimeSeries(topicId, 0, begin, end, end - begin));
        loadTimeSeries(series, begin, end, height, 1, smoothing, isAdaptive);
        filterContradictions(series, threshold, 0.1f, CValues.TYPE_CONTR_HIGH, true);
        return series.get(0);
    }

    public TimeSeries loadTimeSeries(int topicId, int begin, int end, int height, int aggregate, int demographics_id, float smoothing, boolean isAdaptive) {
        HashMap<Integer,TimeSeries> series = new HashMap<Integer,TimeSeries>();
        series.put(demographics_id, new TimeSeries(topicId, demographics_id, begin, end, end - begin));
        loadTimeSeries(series, begin, end, height, aggregate, smoothing, isAdaptive);
        filterContradictions(series, threshold, 0.1f, CValues.TYPE_CONTR_HIGH, false);
        return series.get(demographics_id);
    }

    public void loadTimeSeries(HashMap<Integer,TimeSeries> series, int begin, int end, int height, int aggregate, float smoothing, boolean isAdaptive) {
        //access the time series
        if (!isAdaptive) {
            index.setInterval(height, begin, end);
            for (CPage page : index) {
                for (TimeSeries vc : series.values()) {
                    CValues v = page.getValues(vc.topicId);
                    v.p = 1;
                    vc.add(v, true);
                }
            }
        } else {
            index.setInterval(height + 1, begin, end);
            for (CPage parent_page : index) {
                for (TimeSeries vc : series.values()) {
                    CValues parent_v = parent_page.getValues(vc.topicId);
                    float parent_threshold = parent_v.getContr() / aggregate;
                    for (int i = 0; i < parent_page.numChildren; i++) {
                        if (parent_page.subpages.get(i) > 0 && parent_page.timestamps.get(i) > begin && (parent_page.timestamps.get(i) - parent_page.subInt) < end) {
                            CPage page = (CPage) index.readPage(parent_page.subpages.get(i));
                            CValues v = page.getValues(vc.topicId);
                            v.p = parent_threshold;
                            vc.add(v);
                        }
                    }
                }
            }
        }

        // merging intervals
        for (TimeSeries vc : series.values()) {
            if (aggregate > 1) {
                int i = aggregate;
                ArrayList<CValues> aggregated = new ArrayList<CValues>();
                for (CValues v : vc) {
                    CValues last = aggregated.isEmpty() ? null : aggregated.get(aggregated.size()-1);
                    if (i < aggregate) {
                        if (last != null) { last.update(v); last.end = v.end; } i++;
                    } else {
                        aggregated.add(v); i = 1;
                    }
                }
                vc.clear();
                vc.addAll(aggregated);
            }
            if (!vc.isEmpty()) {
                CValues first = vc.get(0);
                CValues last = vc.get(vc.size()-1);
                vc.window = first.end - first.begin;
                vc.begin = first.begin;
                vc.end = last.end;
            }
        }

    }

    public void filterContradictions(HashMap<Integer,TimeSeries> series, float threshold, float delta_volume, int outputType, boolean replace) {
        // filtering contradictions
        for (TimeSeries vc : series.values()) {
            filterContradictions(vc, threshold, delta_volume, outputType, replace, false);
        }
    }

    public TimeSeries filterContradictions(TimeSeries vc, float threshold, float delta_volume, int outputType, boolean replace, boolean merge) {
        float[] stats = vc.stats(TimeSeries.AVERAGE);
        vc.c = CModel.statsAgainstAvg ? stats[2] : 0;
        float delta = delta_volume * sliding_range * (stats[1] - stats[0]) / 2; //delta volume
        ArrayList<CValues> sliding = new ArrayList<CValues>();
        for (CValues v : vc) {
            v.getContr();
            //v.p = threshold * v.p;
            v.type = (v.c > threshold * v.p) ? CValues.TYPE_CONTR_SYNC : CValues.TYPE_CONTR_ZERO;
        }
        TimeSeries out = vc.clone(false);
        boolean exactTime = true;
        for (CValues v : vc) {
            sliding.add(v);
            if (sliding.size() > sliding_range * 2) {
                double left_volume = 0; double right_volume = 0;
                for (int i = 1; i <= sliding_range; i++) {
                    left_volume += sliding.get(sliding_range - i).getAvg() - vc.c;  // vc.c holds the average level
                    right_volume += sliding.get(sliding_range + i).getAvg() - vc.c; // vc.c holds the average level
                }
                CValues middle = sliding.get(sliding_range);
                if (Math.signum(left_volume) * Math.signum(right_volume) < 0 &&
                    Math.abs(left_volume) > delta &&
                    Math.abs(right_volume) > delta) {
                    if (exactTime) {
                        for (int i = 0; i < sliding_range * 2; i++) {
                            CValues v1 = sliding.get(i);
                            CValues v2 = sliding.get(i+1);
                            float v1f = v1.getAvg() - vc.c;
                            float v2f = v2.getAvg() - vc.c;
                            int time = -1;
                            if (Math.signum(v1f) * Math.signum(v2f) < 0) {
                                //(time - v1.time())/v1f = (v2.time() - time)/v2f
                                //time(1+v1f/v2f) = v1.time()+v2.time()*v1f/v2f
                                //time  = (v1.time()*v2f+v2.time()*v1f)/(v1f+v2f)
                                time = (int)((v1.time() * Math.abs(v2f) + v2.time() * Math.abs(v1f))/(Math.abs(v1f) + Math.abs(v2f)));
                            } else if (Math.signum(v1f) == 0) {
                                time = v1.time();
                            } else if (Math.signum(v2f) == 0) {
                                time = v2.time();
                            }
                            if (time > 0) {
                                float n = (v1.n + v2.n)/2;
                                float var = (v1.getVar()+v2.getVar())/2;
                                CValues v0 = new CValues(time, time-vc.window, time+vc.window, n, 0, n*var);
                                v0.type = CValues.TYPE_CONTR_ASYNC;
                                out.add(v0);
                                while (sliding.size() > sliding_range) sliding.remove(0); // the last is removed later
                                break;
                            }
                        }
                    } else if (middle.type == CValues.TYPE_CONTR_SYNC) {
                        CValues previous = sliding.get(sliding_range - 1);
                        if (previous.type != CValues.TYPE_CONTR_ASYNC) {
                            middle.type = CValues.TYPE_CONTR_ASYNC;
                        } else if (previous.c < middle.c) {
                            middle.type = CValues.TYPE_CONTR_ASYNC;
                            previous.type = CValues.TYPE_CONTR_SYNC;
                        }
                    }
                }
                sliding.remove(0);
            }
        }

        CValues last = null;
        for (CValues v : vc) {
            if ((v.type & outputType) > 0) {
                if (merge && last != null && (last.type & v.type) > 0 && v.begin - last.end < 2 * vc.window) {
                    last.end = v.end;
                } else {
                    out.add(v);
                    if (last != null && last.end - last.begin > 2 * vc.window) {
                        int time = last.time();
                        last.begin = time - vc.window;
                        last.end = time + vc.window;
                    }
                    last = v;
                }
            }
        }
        if (replace) {
            vc.clear();
            vc.addAll(out);
            out = vc;
        }

        return out;
    }

    public float fixNaN(float value) {
        return (Float.isNaN(value) || Float.isInfinite(value)) ? 0 : value;
    }

    public void initBuckets() {
        buckets = new HashMap<Integer,TimeSeries>();
    }

    public void updateBuckets(int topicId, int timestamp, float s) {
        updateBuckets(topicId, timestamp, new Values(1, s, s*s));
    }

    public void updateBuckets(int topicId, int timestamp, Values values) {
        if (timestamp >= indexBegin && timestamp < indexEnd) {
            int i = (timestamp - indexBegin) / minInterval;
            int key = ((i << 16) | topicId);
            TimeSeries bucket = buckets.get(key);
            if (bucket == null) {
                bucket = new TimeSeries(topicId, i, indexBegin + i * minInterval, indexBegin + (i+1) * minInterval, minInterval);
                buckets.put(key, bucket);
            }
            bucket.update(values);
        }
    }

    public int maxTopicId = -1;

    private void getMaxTopicId() {
        maxTopicId = -1;
        for (CValues t : topics) {
            if (t.id > maxTopicId) maxTopicId = t.id;
        }
    }

    List<CValues> topics = new ArrayList<CValues>();

    public void updateTopics() {
        topics.clear();
        if (index != null) {
            List<CValues> allTopics = index.rootPage.getTopics(0);
            for (CValues t : allTopics) {
                topics.add(t);
            }
        }
        getMaxTopicId();
    }
}

