package ctree;

import main.CValues;
import main.Values;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.*;

/**
 * A class representing a node in CTree, which can be read and written as a disk page.
 * Extra values that do not fit in a single disk page are stored in a chain of linked
 * {@link ctree.CPageExtent} pages.
 *
 * @author Mikalai Tsytsarau
 */

public class CPage extends CPageExtent {

    public byte numChildren = (byte) -1;
    public int parent = -1; // reference to the parent page
    public int previous = -1; // reference to the preceding page of same granularity
    public int next = -1; // reference to the next page of same granularity
    protected ArrayList<Integer> timestamps = new ArrayList<Integer>(); // marking ends of children time intervals
    protected ArrayList<Integer> subpages = new ArrayList<Integer>(); // refer to children pages of lower granularity


    /**
     * Creates new page.
	 *
     * @param index reference to index
     * @param pageSize pageSize in bytes
     * @param pageId id of the page
     * @param parentId id of a parent page
     * @param height location in the tree
     * @param begin_timestamp time
     * @param end_timestamp time
     * @return newly created page
	 */
    public CPage(CIndex index, int pageSize, int pageId, int parentId, int height, int begin_timestamp, int end_timestamp) {
        this(index, pageSize, pageId);
        init(parentId, height, begin_timestamp, end_timestamp);
    }

    public CPage(CIndex index, int pageSize, int pageId) {
        super(index, DBPage.TYPE_CTREE, pageSize, pageId);
        init();
    }

    private void init() {
        int head_size = 2 + 5 * int_size;
        int links_size = 1 + 2 * int_size + 2 * numChildren * int_size;
        pageCapacity = (pageSize - head_size - links_size) / getPayloadSize();
        subInt = index.getInterval(height - 1);
    }

    public void init(int parentId, int height, int begin_time, int end_time) {
        init(height, begin_time, end_time);
        int t = begin;
        numChildren = index.getNumChildren(height);
        for (int i = 0; i < numChildren; i++) {
            t += subInt;
            timestamps.add(t);
            subpages.add(-1);
        }
        if (numChildren > 0 && t != end) {
            System.err.println("CPage interval problem: t != end: " + t + "-" + end + "=" + (end - t));
            end = t;
        }
        parent = parentId;
        init();
    }

    public CPage insertChild(int childId) {
        CPage page = (CPage) index.newPage(DBPage.TYPE_CTREE);
        int pageEnd = timestamps.get(childId);
        int pageBegin = pageEnd - subInt;
        page.init(pageId, height - 1, pageBegin, pageEnd);
        page.previous = findNeighbour(childId, -1);
        page.next = findNeighbour(childId, 1);
        subpages.set(childId, page.pageId);
        index.writePage(page);
        index.writePage(this);

        if (page.previous > 0) {
            CPage previousPage = (CPage) index.readPage(page.previous);
            previousPage.next = page.pageId;
            index.writePage(previousPage);
        }
        //TODO: We can look only previous or next, and get another one through link
        if (page.next > 0) {
            CPage nextPage = (CPage) index.readPage(page.next);
            nextPage.previous = page.pageId;
            index.writePage(nextPage);
        }

        return page;
    }

    public CPage getChildAt(int i) {
        return (CPage) index.readPage(subpages.get(i));
    }

    public CPage getChild(int timestamp) {
        return getChild(timestamp, false);
    }

    private CPage getChild(int timestamp, boolean create_new) {
        CPage page = null;
        int childId = findChild(timestamp);
        if (childId >= 0 && childId < numChildren) {
            synchronized (subpages) {
                int pageId = subpages.get(childId);
                if (pageId != -1) {
                    page = (CPage) index.readPage(pageId);
                } else if (create_new && height > 0) {
                    page = insertChild(childId);
                }
            }
        }
        return page;
    }

    public int findNeighbour(int childId, int direction) {
        int neighbourId = -1;
        for (int i = childId + direction; i >= 0 && i < numChildren; i += direction) {
            if (subpages.get(i) > 0) {
                neighbourId = subpages.get(i);
                break;
            }
        }
        if (neighbourId < 0) {
            int nextId = (direction > 0) ? next : previous;
            if (nextId > 0) {
                int nextChildId = (direction > 0) ? -1 : numChildren;
                CPage page = (CPage) index.readPage(nextId);
                if (page != null) {
                    neighbourId = page.findNeighbour(nextChildId, direction);
                }
            }
        }
        return neighbourId;
    }

    /**
	 * Reads the object from a byte array.
	 *
	 * @param buffer page
	 */
    protected void read(ByteBuffer buffer) {
		super.read(buffer);
        parent = buffer.getInt();
        previous = buffer.getInt();
        next = buffer.getInt();
        if (next == 0 || previous == 0 || parent == 0) {
            System.err.println("Dirty reads problem! Page: " + pageId);
        }
        numChildren = buffer.get();
        timestamps.clear();
        for (int i = 0; i < numChildren; i++) {
            timestamps.add(buffer.getInt());
            //if (timestamps.size() > 1 &&
            //   (timestamps.get(timestamps.size() - 1) - timestamps.get(timestamps.size() - 2)) == interval/CPage.numChildren) {
            //    System.out.println(pageId + " " + (timestamps.get(timestamps.size() - 1) - timestamps.get(timestamps.size() - 2)) + " " + interval/CPage.numChildren);
            //}
        }
        subpages.clear();
        for (int i = 0; i < numChildren; i++) {
            subpages.add(buffer.getInt());
        }
        init();
    }
    /*
    protected void read(ByteBuffer buffer) {
        pageType = (byte) buffer.getInt();
        height = (byte) buffer.getInt();
        interval = buffer.getInt();
        end = (int) buffer.getLong();
        begin = end - interval;

        parent = -1;
        previous = buffer.getInt();
        next = buffer.getInt();
            int numChilds = buffer.getInt();
            timestamps.clear();
            for ( int i=0; i<numChilds; i++ ) {
                timestamps.add((int)buffer.getLong());
            }
        subpages.clear();
            for ( int i = 0; i < numChilds; i++ ) {
                subpages.add(buffer.getInt());
            }

        pageExtent = buffer.getInt();
        int numTopics = buffer.getInt();
        for ( int i = 0; i < numTopics; i++ ) {
            topics.add(buffer.getInt());
            vals.add(new Values(buffer.getInt(), buffer.getFloat(), buffer.getFloat()));
        }
        init();
    }
     */
    /**
	 * Writes the object to a byte array.
	 *
	 * @param buffer page
	 */
	protected void write(ByteBuffer buffer) {
		super.write(buffer);
        buffer.putInt(parent);
        buffer.putInt(previous);
        buffer.putInt(next);
        byte numChildren = (byte) timestamps.size();
        buffer.put(numChildren);
        for (int i = 0; i < numChildren; i++) {
            buffer.putInt(timestamps.get(i));
        }
        for (int i = 0; i < numChildren; i++) {
            buffer.putInt(subpages.get(i));
        }
    }

    public void sort() {
        //first sort individual pages, then merge the result
        List<List<CValues>> topics = new ArrayList<List<CValues>>();
        //stores current positions in each List<CTopic> array
        //List<Integer> positions = new ArrayList<Integer>();
        List<CPageExtent> pages = new ArrayList<CPageExtent>();
        CPageExtent page = this;
        while (page != null) {
            List<CValues> page_topics = page.getPageValues();
            Collections.sort(page_topics);
            topics.add(page_topics);
            pages.add(page);
            //System.out.println(page + " topics:" + page_topics.size());
            page = (CPageExtent) index.readPage(page.pageExtent);
        }
        //System.out.flush();
        //current page to write to
        page = this;
        //position in the current page
        int pos = 0; int page_pos = 0;
        //merge the result
        while (topics.size() > 0) {
            //find max
            float max_ct = topics.get(0).get(0).c;
            int max_i = 0;
            for (int i = 1; i < topics.size(); i++) {
                float ct = topics.get(i).get(0).c;
                if (ct > max_ct) {
                    max_ct = ct;
                    max_i = i;
                }
            }
            //write max to the place
            page.setValuesAt(page_pos, topics.get(max_i).get(0));
            //System.out.printf(page_pos + ",");

            topics.get(max_i).remove(0);
            if (topics.get(max_i).isEmpty()) {
                topics.remove(max_i);
            }

            //update positions
            page_pos++;
            if (page_pos >= page.size()) {
                index.writePage(page);
                pos++;
                page_pos = 0;
                //System.out.printf("\n%s topics:%d\n", page.toString(), page.topicIds.size());
                //System.out.flush();
                if (pos < pages.size()) {
                    page = pages.get(pos);
                } else {
                    break;
                }
            }
        }
        //tree.writePage(page);
        //continue with all children
        for (int i = 0; i < numChildren; i++) {
            CPage subpage = (CPage) index.readPage(this.subpages.get(i));
            if (subpage != null) {
                subpage.sort();
            }
        }
    }

     /**
     * Inserts the given sentiment (updating topic or adding new one).
     * Not working for large scale, since adding +1 doesn't change large numbers
     * Use batch inserts for the large scale instead
     *
     * @param topicId tid
     * @param timestamp timestamp
     * @param s Sentiment.
     * @return updated topic
     */
    public Values insert(int topicId, int timestamp, float s) {
        return insert(topicId, timestamp, new Values(1, s, s * s), -1)[1];
    }

    /**
     * Inserts the given sentiment (updating topic or adding new one).
     * Not working for large scale, since adding +1 doesn't change large numbers
     * Use batch inserts for the large scale instead
     * If compressed, and not using demographics as raw storage, then
     * leaf nodes are updated as raw, updates are propagated to parents (compressed)
     *
     * @param topicId tid
     * @param timestamp timestamp
     * @param values values to update/insert
     * @param demographics_id demographics
     * @return array of values (old,updated)
     */
    public Values[] insert(int topicId, int timestamp, Values values, int demographics_id) {
        if (height > 0) {
            CPage child = getChild(timestamp, true);
            if (child != null) {
                Values[] upd = child.insert(topicId, timestamp, values, demographics_id);
                if (index.compressed && !index.extended) {
                    upd[0].normalize(); upd[1].normalize();
                    return updateValues(topicId, new Values(upd[1].n - upd[0].n, upd[1].m1 - upd[0].m1, upd[1].m2 - upd[0].m2), 0); //remove old//add new
                }
            } else {
                System.err.println("Child not found: " + timestamp + " (" + begin + "," + end + ")");
            }
        }
        return updateValues(topicId, values, demographics_id);
    }

    public int findChild(int timestamp) {
        int j = 0;
        if (timestamp <= begin) {
            j = -1;
        } else if (timestamp > end) {
            j = numChildren;
        } else {
            while (timestamps.get(j) < timestamp) {
                j++;
            }
        }
        return j;
    }

    public ArrayList<CPage> getSubPages() throws IOException {
        //list all sub-intervals
        ArrayList<CPage> subPages = new ArrayList<CPage>();
        for (Integer i : subpages) {
            CPage page = (CPage) index.readPage(i);
            if (page != null) {
                subPages.add(page);
            }
        }
        return subPages;
    }

}