package ctree;

import main.CValues;
import main.Values;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.List;

/**
 * A class for a disk page representing a {@link ctree.CPage} extent node in CTree.
 *
 * @author Mikalai Tsytsarau
 */

@SuppressWarnings({"ConstantConditions"})
public class CPageExtent extends DBPage {
    public byte height;  //height (granularity) of the current TPage (zero is leaf page)
    public int begin = -1; // begin of time interval
    public int end = -1; // end of time interval
    public int pageExtent = -1; // additional page for storing topics (optional)
    public int subInt = -1;

    protected ArrayList<Integer> topics = new ArrayList<Integer>(); // a list of topics allocated in that page
    protected ArrayList<Values> vals = new ArrayList<Values>();

    public CPageExtent(CIndex index, int pageSize, int pageId) {
        this(index, DBPage.TYPE_CTREE_EXTENT, pageSize, pageId);
    }

    public CPageExtent(CIndex index, int pageType, int pageSize, int pageId) {
        super(index, pageType, pageSize, pageId);
        init();
    }

    public void init(int height, int begin_timestamp, int end_timestamp) {
        this.height = (byte) height;
        this.begin = begin_timestamp;
        this.end = end_timestamp;
        init();
    }

    private void init() {
        int head_size = 2 + 4 * int_size;
        pageCapacity = (pageSize - head_size) / getPayloadSize();
        subInt = index.getInterval(height - 1);
    }

    public long interval() {
        return (long)end - (long)begin;
    }

    public int getPayloadSize() {
        return 1 * int_size + 3 * (index.compressed && height > 0 ? byte_size : float_size);
    }

    public String toString() {
        java.text.SimpleDateFormat stub = new java.text.SimpleDateFormat("dd.MM.yyyy");
        return stub.format(new java.util.Date(begin * 1000L)) + "-"
                + stub.format(new java.util.Date(end * 1000L));
                //+ " h:" + height + " ext:" + topicExtent;
    }

    public Values[] updateValues(CValues values, int demographics_id) {
        return updateValues(values.id, values, demographics_id);
    }

    /**
     * Inserts a new topic into this page.
     * @param topicId  the topicId
     * @param values  the n, m1, m2
     * @param demographics_id  default -1
     * @return array of (old,updated) values
     */
    public Values[] updateValues(int topicId, Values values, int demographics_id) {
        index.writePage(this); // we update pageExtent in the middle - have to mark as dirty
        Values[] result = new Values[2];
        synchronized (topics) {
            int i = topics.indexOf(topicId);
            if (i < 0) {
                if (pageExtent <= 0 && this.size() < pageCapacity) {
                    // insert new topic here
                    i = this.size();
                    topics.add(topicId);
                    vals.add(new Values());
                } else { // still not found?
                    CPageExtent extent;
                    if (pageExtent > 0) {
                        extent = (CPageExtent) index.readPage(pageExtent);
                    } else {
                        extent = (CPageExtent) index.newPage(DBPage.TYPE_CTREE_EXTENT);
                        extent.init(height, begin, end);
                        pageExtent = extent.pageId;
                    }
                    return extent.updateValues(topicId, values, demographics_id);
                }
            }

            if (i >= 0) {
                Values old = vals.get(i);
                result[0] = old.clone();
                old.update(values);
                result[1] = old.clone();
            }

        }
        return result;
    }

    protected boolean setValuesFor(int topicId, CValues values) {
        synchronized (topics) {
            int id = topics.indexOf(topicId);
            if(id >= 0) {
                setValuesAt(id, values);
                return true;
            } else {
                if (pageExtent > 0) {
                    CPageExtent extent = (CPageExtent) index.readPage(pageExtent);
                    return extent.setValuesFor(topicId, values);
                } else {
                    return false;
                }
            }
        }
    }

    protected void setValuesAt(int i, CValues values) {
        synchronized (topics) {
            topics.set(i, values.id);
            vals.set(i, values);
        }
    }

    // for testing scalability
    public void replicateValues(CPageExtent previous, int delay, int maxTopics) {
        CPageExtent extent;
        if (delay > pageCapacity || (previous == null && delay > this.size())) {
            return;
        }

        if (pageExtent > 0) {
            //go to the last page
            extent = (CPageExtent) index.readPage(pageExtent);
            extent.replicateValues(this, delay, maxTopics);
            return;
        }

        int lastTopic = 0;
        if (!topics.isEmpty()) {
            lastTopic = topics.get(topics.size() - 1);
        } else if (previous != null) {
            lastTopic = previous.topics.get(previous.topics.size() - 1);
        }

        while (this.size() < pageCapacity && lastTopic < (maxTopics - 1)) {
            lastTopic++;
            int source = this.size() - delay;
            if (source >= 0) {
                vals.add(vals.get(source));
            } else if (previous != null) {
               source += previous.size();
               vals.add(previous.vals.get(source));
            }
            topics.add(lastTopic);
            //n.add((int) Math.round(Math.random() * 10 + 1));
            //m1.add((float) (Math.random() - 0.5) * n.lastElement());
            //m2.add((float) (Math.random() * 0.5 * n.lastElement()));
        }
        if (lastTopic < (maxTopics - 1)) {
            extent = (CPageExtent) index.newPage(DBPage.TYPE_CTREE_EXTENT);
            extent.init(height, begin, end);
            pageExtent = extent.pageId;
            extent.replicateValues(this, delay, maxTopics);
        }
        index.writePage(this);
    }

    private CValues getValuesAt(int i) {
        CValues topic = new CValues(topics.get(i), begin, end);
        Values values = vals.get(i);
        if (values != null) topic.update(values);
        return topic;
    }

    public CValues getValues(int topicId) {
        CValues values;
        int id = topics.indexOf(topicId);
        if(id >= 0) {
            values = getValuesAt(id);
        } else {
            if (pageExtent > 0) {
                CPageExtent extent = (CPageExtent) index.readPage(pageExtent);
                values = extent.getValues(topicId);
            } else {
                values = new CValues(topicId, begin, end);
            }
        }
        return values;
    }

    public CValues getValuesTopK(int topicId, float minValue) {
        if (getValuesAt(0).c < minValue) {
            return null;
        }
        CValues values = null;
        int id = topics.indexOf(topicId);
        if(id >= 0) {
            values = getValuesAt(id);
            if (values.c < minValue) {
                values = null;
            }
        } else {
            if (pageExtent > 0) {
                CPageExtent extent = (CPageExtent) index.readPage(pageExtent);
                values = extent.getValuesTopK(topicId, minValue);
            }
        }
        return values;
    }

    public List<CValues> getAllValues() {
        return getAllValues(0);
    }

    public List<CValues> getAllValues(int page_limit) {
        List<CValues> values = new ArrayList<CValues>();
        CPageExtent extent = this;
        do {
            for (int i = 0; i < extent.size(); i++) {
                values.add(extent.getValuesAt(i));
            }
            page_limit--;
            if (page_limit != 0) {
                extent = (CPageExtent) index.readPage(extent.pageExtent);
            } else {
                extent = null;
            }
        } while (extent != null);
        return values;
    }

    public List<CValues> getPageValues() {
        List<CValues> values = new ArrayList<CValues>();
        for (int i = 0; i < this.size(); i++) {
            values.add(getValuesAt(i));
        }
        return values;
    }

    public List<CValues> getTopics(int topic_limit) {
        List<CValues> topics = new ArrayList<CValues>();
        CPageExtent extent = this;
        do {
            for (int i = 0; i < extent.size(); i++) {
                topics.add(extent.getValuesAt(i));
                topic_limit--;
                if (topic_limit == 0) break;
            }
            if (topic_limit != 0) {
                extent = (CPageExtent) index.readPage(extent.pageExtent);
            } else {
                extent = null;
            }
        } while (extent != null);
        return topics;
    }

    public int size() {
        return topics.size();
    }
     /**
     * Reads the object from a byte array.
     *
     * @param buffer page
     */
    protected void read(ByteBuffer buffer) {
        pageType = buffer.get();
        height = buffer.get();
        begin = buffer.getInt();
        end = buffer.getInt();
        //OLD INDEX: begin contains window size
        if (CModel.indexVersion == CModel.INDEX_VER_0) {
            begin = end - begin;
        }
        pageExtent = buffer.getInt();
        int numTopics = buffer.getInt();

        topics.clear();
        vals.clear();
        for (int i = 0; i < numTopics; i++) {
            topics.add(buffer.getInt());
            if (!index.compressed || height == 0) {
                vals.add(new Values(buffer.getFloat(), buffer.getFloat(), buffer.getFloat()));
            } else {
                vals.add(new Values(buffer.get(), buffer.get() / 12.7f, (buffer.get() + 127) / 25.5f));
            }
        }
        init();
    }
    /*
    protected void read(ByteBuffer buffer) {
        pageType = (byte) buffer.getInt();
        height = (byte) buffer.getInt();
        interval = buffer.getInt();
        end = (int) buffer.getLong();
        begin = end - interval;

        pageExtent = buffer.getInt();
        int numTopics = buffer.getInt();
        for ( int i = 0; i < numTopics; i++ ) {
            topics.add(buffer.getInt());
            vals.add(new Values(buffer.getInt(), buffer.getFloat(), buffer.getFloat()));
        }
        init();
    }
    */
    /**
     * Writes the object to a byte array.
     *
     * @param buffer page
     */
    protected void write(ByteBuffer buffer) {
        buffer.put(pageType);
        buffer.put(height);
        buffer.putInt(begin);
        buffer.putInt(end);
        buffer.putInt(pageExtent);
        int numTopics = vals.size();
        buffer.putInt(numTopics);

        for (int i = 0; i < numTopics; i++) {
            buffer.putInt(topics.get(i));
            if (!index.compressed) {
                buffer.putFloat(vals.get(i).n);
                buffer.putFloat(vals.get(i).m1);
                buffer.putFloat(vals.get(i).m2);
            } else {
                buffer.put((byte) vals.get(i).n);
                buffer.put((byte) (vals.get(i).m1 * 12.7f));
                buffer.put((byte) (vals.get(i).m2 * 25.5f - 127));
            }
        }
    }

}
