package ctree;

/**
 * A class for managing database-based and retrieving contradictions.
 * It can read, write and update sentiment values from a database table via {@link ctree.Model} interface.
 *
 * @author Mikalai Tsytsarau;
 */

import database.DatabaseAccess;
import main.CValues;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

public class DBModel implements Model {
    int databaseType = -1;

    public DBModel() {
        if (DatabaseAccess.initialized()) {
            try {
                String database = DatabaseAccess.getConnection().getMetaData().getDatabaseProductName();
                if (database.equalsIgnoreCase("MySQL")) {
                    databaseType = 0;
                } else {
                    databaseType = 1;
                }
            } catch (SQLException e) {
                e.printStackTrace();
            }
        }
    }

    public List<CValues> getContradictingTopics(int begin, int end, int height, float threshold, boolean isRecursive, int limit) {
        List<CValues> topics = new ArrayList<CValues>();
        ResultSet rs = null;
        String query = "";
        int num_limit = limit * 244;
        try {
            if (!isRecursive) {
                query = "select distinct c1.topic_id from contradictions c1 \n" +
                        "where (c1.timeBegin between @timeBegin and @timeEnd \n" +
                        "or c1.timeEnd between @timeBegin and @timeEnd)\n" +
                        "and c1.granularity = @granularity\n" +
                        "and (c1.topic_id < @limit or @limit = 0)\n" +
                        "and contradict(c1.n, c1.m1, c1.m2) > @threshold";
            } else {
                query = "select distinct c1.topic_id from contradictions c1 \n" +
                        "join contradictions c2 \n" +
                        "on c1.topic_id = c2.topic_id \n" +
                        "and (c1.timeBegin between c2.timeBegin and c2.timeEnd)\n" +
                        "and c2.granularity = c1.granularity + 1\n" +
                        "where (c1.timeBegin between @timeBegin and @timeEnd \n" +
                        "or c1.timeEnd between @timeBegin and @timeEnd)\n" +
                        "and c1.granularity = @granularity\n" +
                        "and (c1.topic_id < @limit or @limit = 0)\n" +
                        "and contradict(c1.n, c1.m1, c1.m2) > contradict(c2.n, c2.m1, c2.m2) * @threshold";
            }

            query = query.replaceAll("@timeBegin", String.valueOf(begin));
            query = query.replaceAll("@timeEnd", String.valueOf(end));
            query = query.replaceAll("@granularity", String.valueOf(height));
            query = query.replaceAll("@limit", String.valueOf(num_limit));
            query = query.replaceAll("@threshold", String.valueOf(threshold));
/*
            try {
                debugOut.write(query);
                debugOut.newLine();
            } catch (IOException e) {
                e.printStackTrace();
            }
            */
            rs = DatabaseAccess.executeQuery(query);

            while (rs.next()) {
                topics.add(new CValues(rs.getInt(1), begin, end));
            }
            rs.close();

        }
        catch (SQLException ex) {
            System.out.println("SQLException: " + ex.getMessage());
            System.out.println("SQLState: " + ex.getSQLState());
            System.out.println("VendorError: " + ex.getErrorCode());
            System.out.println("Query: " + query);
        }
        return topics;
    }

    public List<CValues> getContradictingTopicsTopK(int begin, int end, int height, int k) {
        List<CValues> topics = new ArrayList<CValues>();
        ResultSet rs = null;
        String query = "";
        try {
            query = "select * from contradictions c1 \n" +    //distinct c1.topic_id
                        "where (c1.timeBegin between @timeBegin and @timeEnd \n" +
                        "or c1.timeEnd between @timeBegin and @timeEnd)\n" +
                        "and c1.granularity = @granularity\n" +
                        "order by contradict(c1.n, c1.m1, c1.m2) desc\n" +
                        (databaseType == 0 ? "limit @k\n" : "FETCH FIRST @k ROWS ONLY\n" );


            query = query.replaceAll("@timeBegin", String.valueOf(begin));
            query = query.replaceAll("@timeEnd", String.valueOf(end));
            query = query.replaceAll("@granularity", String.valueOf(height));
            query = query.replaceAll("@k", String.valueOf(k));
/*
            try {
                debugOut.write(query);
                debugOut.newLine();
            } catch (IOException e) {
                e.printStackTrace();
            }
            */
            rs = DatabaseAccess.executeQuery(query);

            while (rs.next()) {
                topics.add(new CValues(rs.getInt(1), rs.getInt(2), rs.getInt(3), rs.getInt(5), rs.getFloat(6), rs.getFloat(7)));
                //topics.add(new CTopic(rs.getInt(1), begin, end, height, 0, 0, 0));
            }
            rs.close();

        }
        catch (SQLException ex) {
            System.out.println("SQLException: " + ex.getMessage());
            System.out.println("SQLState: " + ex.getSQLState());
            System.out.println("VendorError: " + ex.getErrorCode());
            System.out.println("Query: " + query);
        }
        return topics;
    }

    public List<CValues> getContradictions(int topicId, int begin, int end, int height, float threshold, boolean isRecursive) {
        List<CValues> topics = new ArrayList<CValues>();
        ResultSet rs = null;
        String query = "";

        try {
            if (!isRecursive) {
                query = "select c1.topic_id, c1.timeBegin, c1.timeEnd, c1.granularity, c1.n, c1.m1, c1.m2, contradict(c1.n, c1.m1, c1.m2) as contradiction from contradictions c1 \n" +
                        "where c1.topic_id = @tid\n" +
                        "and (c1.timeBegin between @timeBegin and @timeEnd \n" +
                        "or c1.timeEnd between @timeBegin and @timeEnd)\n" +
                        "and c1.granularity = @granularity\n" +
                        "and contradiction > @threshold";
            } else {
                query =
                        "select c1.* from contradictions c1 \n" +
                        "join contradictions c2 \n" +
                        "on c1.topic_id = c2.topic_id \n" +
                        "and (c1.timeBegin between c2.timeBegin and c2.timeEnd)\n" +
                        "and c2.granularity = c1.granularity + 1\n" +
                        "where c1.topic_id = @tid\n" +
                        "and (c1.timeBegin between @timeBegin and @timeEnd \n" +
                        "or c1.timeEnd between @timeBegin and @timeEnd)\n" +
                        "and c1.granularity = @granularity\n" +
                        "and contradict(c1.n, c1.m1, c1.m2) > contradict(c2.n, c2.m1, c2.m2) * @threshold";
            }
            query = query.replaceAll("@tid", String.valueOf(topicId));
            query = query.replaceAll("@timeBegin", String.valueOf(begin));
            query = query.replaceAll("@timeEnd", String.valueOf(end));
            query = query.replaceAll("@granularity", String.valueOf(height));
            query = query.replaceAll("@threshold", String.valueOf(threshold));
/*
            try {
                debugOut.write(query);
                debugOut.newLine();
            } catch (IOException e) {
                e.printStackTrace();
            }
            */
            rs = DatabaseAccess.executeQuery(query);

            while (rs.next()) {
                topics.add(new CValues(rs.getInt(1), rs.getInt(2), rs.getInt(3), rs.getInt(5), rs.getFloat(6), rs.getFloat(7)));
            }
            rs.close();

        }
        catch (SQLException ex) {
            System.out.println("SQLException: " + ex.getMessage());
            System.out.println("SQLState: " + ex.getSQLState());
            System.out.println("VendorError: " + ex.getErrorCode());
            System.out.println("Query: " + query);
        }
        return topics;
    }

    public List<CValues> getContradictionsTopK(int topicId, int begin, int end, int height, int k) {
        List<CValues> topics = new ArrayList<CValues>();
        ResultSet rs = null;
        String query = "";

        try {
            query = "select c1.topic_id, c1.timeBegin, c1.timeEnd, c1.granularity, c1.n, c1.m1, c1.m2, contradict(c1.n, c1.m1, c1.m2) as contradiction from contradictions c1 \n" +
                    "where c1.topic_id = @tid\n" +
                    "and (c1.timeBegin between @timeBegin and @timeEnd \n" +
                    "or c1.timeEnd between @timeBegin and @timeEnd)\n" +
                    "and c1.granularity = @granularity\n" +
                    "order by contradiction desc\n" +
                    (databaseType == 0 ? "limit @k\n" : "FETCH FIRST @k ROWS ONLY\n" );

            query = query.replaceAll("@tid", String.valueOf(topicId));
            query = query.replaceAll("@timeBegin", String.valueOf(begin));
            query = query.replaceAll("@timeEnd", String.valueOf(end));
            query = query.replaceAll("@granularity", String.valueOf(height));
            query = query.replaceAll("@k", String.valueOf(k));
/*
            try {
                debugOut.write(query);
                debugOut.newLine();
            } catch (IOException e) {
                e.printStackTrace();
            }
            */
            rs = DatabaseAccess.executeQuery(query);

            while (rs.next()) {
                topics.add(new CValues(rs.getInt(1), rs.getInt(2), rs.getInt(3), rs.getInt(5), rs.getFloat(6), rs.getFloat(7)));
            }
            rs.close();

        }
        catch (SQLException ex) {
            System.out.println("SQLException: " + ex.getMessage());
            System.out.println("SQLState: " + ex.getSQLState());
            System.out.println("VendorError: " + ex.getErrorCode());
            System.out.println("Query: " + query);
        }
        return topics;
    }
}
