package ctree;

import java.io.IOException;
import java.nio.ByteBuffer;

/**
 * Abstract base class for a memory-version disk page.
 * For example, a derived class can be <code>BTreeIndexPage</code>.
 * A derived class should implement the functions to <code>read</code> from
 * a byte array, and to <code>write</code> to a byte array. 
 * A convention is that in the byte array, the first integer should be nodeType.
 * 
 * Created Tian Xia &lt;tianxia@ccs.neu.edu&gt;<br>Donghui Zhang &lt;donghui@ccs.neu.edu&gt;
 * @author Mikalai Tsytsarau
 */
public abstract class DBPage {

    public static final int TYPE_ARRAY = -1;
    public static final int TYPE_EMPTY = 0;
    public static final int TYPE_CTREE = 1;
    public static final int TYPE_CTREE_EXTENT = 2;


    public static final int int_size = Integer.SIZE / 8;
    public static final int byte_size = Short.SIZE / 8;
    public static final int float_size = Float.SIZE / 8;
    public static final int short_size = Short.SIZE / 8;

    /**
	 * The page size.
	 */
	public static int pageSize;

    public int pageCapacity = -1;

	/**
	 * The page type. -1 means a byte array page.
	 * 0 means an empty page. 
	 * Users can define their own node types.
	 */
	public byte pageType;

    /**
	 * The page id.
	 */
	public int pageId;

    CIndex index = null;

	/**
	 * Creates a DBPage.
	 * @param pageType node type
	 * @param pageSize page size
	 */
	public DBPage(CIndex index, int pageType, int pageSize, int pageId ) {
		this.index = index;
        this.pageType = (byte) pageType;
		this.pageSize = pageSize;
        this.pageId = pageId;
    }
	
	/**
	 * Reads the object from a byte array.
	 * 
	 * @param buffer
	 * @throws IOException
	 */
	protected abstract void read(ByteBuffer buffer);

	/**
	 * Writes the object to a byte array.
	 * 
	 * @param buffer
	 * @throws IOException
	 */
	protected abstract void write(ByteBuffer buffer);
}

/**
 * An empty page.
 * All empty pages are linked together.
 * @author Donghui Zhang &lt;donghui@ccs.neu.edu&gt;
 */
class DBEmptyPage extends DBPage {
	/**
	 * pageID of the next empty page; -1 if none.
	 */
	public int nextEmptyPage;
	
	/**
	 * Constructor of <code>DBEmptyPage</code>.
	 * @param pageSize      page size
	 * @param nextEmptyPage the next empty page
	 */
	public DBEmptyPage(CIndex index, int pageSize, int pageId, int nextEmptyPage) {
		super(index, 0, pageSize, pageId);
		this.nextEmptyPage = nextEmptyPage;
	}

    public DBEmptyPage(CIndex index, int pageSize, int pageId) {
		super(index, 0, pageSize, pageId);
	}

	/**
	 * Reads the second integer from the beginning of the page.
	 * This is the pointer to the next empty page.
	 */
	protected void read(ByteBuffer buffer) {
        buffer.get();
        nextEmptyPage = buffer.getInt();
	}

	/**
	 * Write two integers to the beginning of the page.
	 * The first integer is 0 standing for empty page.
	 * The second integer is a pointer to the next empty page.
	 * Note that this is the last empty page, the pointer is -1. 
	 */
	protected void write(ByteBuffer buffer) {
		buffer.put((byte) 0);
        buffer.putInt(nextEmptyPage);
	}
}

/**
 * A <code>DBPage</code> which contains a byte array.  
 * An example usage of <code>DBByteArrayPage</code> is to use it as the header page
 * of the file. It is crucial that before calling {@link CIndex#writePage}
 * to write a <code>DBByteArrayPage</code> to buffer,
 * the application writes its header information to its stored
 * byte array: <code>content</code>.
 * 
 * @author Donghui Zhang &lt;donghui@ccs.neu.edu&gt;
 */
class DBByteArrayPage extends DBPage {
	/**
	 * the stored byte array
	 */
	protected byte[] content;
	/**
	 * Constructor.
	 * @param pageSize page size
	 * @param buffer the source byte[] to be copied to <code>content</code>
	 */
	public DBByteArrayPage(CIndex index, int pageSize, int pageId, ByteBuffer buffer) {
		super(index, -1, pageSize, pageId);
        content = new byte[pageSize];
        read(buffer);
	}

    public DBByteArrayPage(CIndex index, int pageSize, int pageId) {
		super(index, -1, pageSize, pageId);
        content = new byte[pageSize];
	}

	protected void read(ByteBuffer buffer) {
        buffer.position(0);
        buffer.get(content);
		//System.arraycopy(buf, 0, content, 0, pageSize);
        //content = buffer.array();
        //does not work for direct allocations
	}

	protected void write(ByteBuffer buffer) {
        buffer.position(0);
        if (buffer.remaining() >= content.length) {
            buffer.put(content);
        } else {
            System.err.println("Buffer size smaller than array!");
            System.err.println("Remaining: " + buffer.remaining() + " Array: " + content.length);
        }
	}
}