package ctree;

import java.io.*;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.file.*;
import java.util.ArrayList;
import java.util.Collections;

/**
 * Abstract base class for a buffer manager.
 * It helps index structures (derived from {@link CIndex}) manage disk pages.
 * <p>
 * <b>Usage:</b><br>
 * <ul>
 * <li> The user can {@link #pin} or {@link #unpin} a page.
 * <li> The user can call {@link #getIOs} to get statistics, or call {@link #clearIOs}
 * to initialize the counters.
 * </ul>
 *
 * @see CIndex
 * @see CPage
 * Created Tian Xia &lt;tianxia@ccs.neu.edu&gt;<br>Donghui Zhang &lt;donghui@ccs.neu.edu&gt;
 * @author Mikalai Tsytsarau
 */
public abstract class LRUBuffer {

    protected ArrayList<CBufferElement> pageBuffer;

    //underlying file
    public String fileName;
    private RandomAccessFile f = null;
    private FileDescriptor fd = null;
    private FileChannel file = null;
    /**
     * number of read IOs on the buffer
     */
    protected int bufferReadIO;
    /**
     * number of write IOs on the buffer
     */
    protected int bufferWriteIO;
    /**
     * number of read IOs on the disk
     */
    protected int diskReadIO;
    /**
     * number of write IOs on the disk
     */
    protected int diskWriteIO;
    /**
     * how many pages the buffer has
     */
    protected int bufferSize;
    /**
     * page size
     */
    public int pageSize;
    private boolean dirty = false;

    /**
     * Opens an existing Database file.
     *
     * @param file file to store sequence of pages
     * @param bufferSize	total number of buffer pages
     * @param pageSize		page size
     */
    public LRUBuffer(File file, int bufferSize, int pageSize) {
        if (file != null) {
            try {
                this.f = new RandomAccessFile(file, "rws");
                this.fd = this.f.getFD();
                this.file = f.getChannel();
                this.fileName = file.getPath(); //File.separator is also ok
            } catch (IOException e) {
                e.printStackTrace();
            }
        } else {
            this.fileName = "memory buffer";
        }
        this.bufferSize = bufferSize;
        this.pageSize = pageSize;
        pageBuffer = new ArrayList<CBufferElement>();
        clearIOs();
    }

    /**
     * Resets the four counters <code>bufferReadIO</code>, etc.
     */
    public void clearIOs(){
        bufferReadIO = 0;
        bufferWriteIO = 0;
        diskReadIO = 0;
        diskWriteIO = 0;
    }

    /**
     * Reports the disk/buffer I/Os. There are four I/Os reported:
     * <i>disk-read</i>, <i>disk-write</i>, <i>buffer-read</i> and
     * <i>buffer-write</i>. They are stored in the integer array in
     * the above order.
     *
     * @return  the four I/Os
     */
    public int[] getIOs() {
        int[] io = new int[4];
        io[0] += diskReadIO;
        io[1] += diskWriteIO;
        io[2] += bufferReadIO;
        io[3] += bufferWriteIO;
        return io;
    }

    public void resize(int bufferSize) throws IOException{
        synchronized (pageBuffer) {
            this.bufferSize = bufferSize;
            int diff = pageBuffer.size() - bufferSize;

            ArrayList<CBufferElement> removed = new ArrayList<CBufferElement>();

            for (CBufferElement e : pageBuffer) {
                if (removed.size() < diff) {
                    if (e.pinCount == 0) {
                        removed.add(e);
                    }
                } else {
                    break;
                }
            }
            System.out.println("resize " + fileName + " pages: " + removed.size());
            for (CBufferElement e : removed) {
                pageBuffer.remove(e);
                savePage(e);
            }
            sync();
            assert pageBuffer.size() <= bufferSize : "Buffer full! Pages left after resize!";
        }
    }

    /**
     * Loads a page (the byte array) from disk.
     * Increases <code>diskReadIO</code> by one.
     *
     * @param page    the byte array of length pageSize
     * @param pageID  the page ID
     * @throws IOException
     */
    protected void load(ByteBuffer page, int pageID) throws IOException {
        if (file == null) return;
        int size = file.read(page, pageSize*pageID);
        if (size != pageSize) {
            throw new IOException("Loading a page < page size! File=" + file + "; pageID=" + pageID + "; Size=" + size + ";");
        }
        page.position(0);
        diskReadIO++;
    }

    /**
     * Saves a page (the byte array) to the disk.
     * Increases <code>diskWriteIO</code> by one.
     *
     * @param page   the byte array of length pageSize
     * @param pageID the page ID
     * @throws IOException
     */
    protected void save(ByteBuffer page, int pageID) throws IOException {
        if (file == null) return;
        page.position(0);
        /*
        if (page.capacity() != pageSize) {
            System.out.println( "Plans to save a page whose size is not pageSize!" );
            System.exit(0);
        }
        */
        dirty = true;
        diskWriteIO++;
        file.write(page, pageSize*pageID);
    }

    /**
    * Finds a page in buffer.
    * If found, move to the beginning of buffer.
    * Note: do not increase bufferReadIO.
    *
    * @param pageID    the page ID
    * @return an instance of DBBufferStoredElement if found; <i>null</i>otherwise.
    */
    protected CBufferElement find(int pageID) {
        synchronized (pageBuffer) {
            for (CBufferElement s : pageBuffer) {
                if (s.pageID == pageID) {
                    return s;
                }
            }
        }
        return null;
    }

    /**
    * Adds a page to the buffer.
    * Note: do not increase bufferReadIO.
    * The element should not exist in the buffer already.
    *
    * @param page the stored element
    * @throws java.io.IOException if bad happens
    */
    protected void add(CBufferElement page) throws IOException{
        synchronized (pageBuffer) {
            if (pageBuffer.size() >= bufferSize) {
                // find a page to switch out
                CBufferElement e = null;
                for (CBufferElement s : pageBuffer) {
                    if (s.pinCount == 0) {
                        e = s;
                        break;
                    }
                }
                assert e != null : "Buffer full! No page to switch out!";
                pageBuffer.remove(e);

                // write to disk if s is dirty
                savePage(e);
                sync();
            }
            pageBuffer.add(page);
        }
    }

    /**
     * Flushes the buffer pages of a given file.
     * Makes the buffer empty, while saving modified pages to disk.
     *
     * @throws IOException if bad happens
     */
    public void flush() throws IOException {
        synchronized (pageBuffer) {
            for (CBufferElement page : pageBuffer) {
                savePage(page);
            }
            //System.out.println("flush " + fileName + " pages: " + pageBuffer.size());
            pageBuffer.clear();
            sync();
        }
    }

    public void flush(int maxMemorySize, int numPages) throws IOException {
        if (pageBuffer.size() > maxMemorySize) flushLRU(numPages);
    }

    public void flushLRU(int numPages) throws IOException {
        synchronized (pageBuffer) {
            Collections.sort(pageBuffer);
            ArrayList<CBufferElement> removed = new ArrayList<CBufferElement>();
            for (CBufferElement e : pageBuffer) {
                if (removed.size() < numPages) {
                    if (e.pinCount == 0) {
                        removed.add(e);
                    }
                } else {
                    break;
                }
            }
            System.out.println("flush " + fileName + " pages: " + removed.size());
            for (CBufferElement e : removed) {
                pageBuffer.remove(e);
                savePage(e);
            }
            sync();
        }
    }

    //without this page can be read *before* it is written!
    private void sync() throws IOException {
        if (file == null) return; // || !dirty
        file.force(true); fd.sync();
        dirty = false;

        /*
        //create a WatcherService object for monitoring directories
        WatchService watchService = FileSystems.getDefault().newWatchService();
        Path myFile = Paths.get(fileName);
        Path dir = myFile.getParent();
        //register the events to be notified by the program
        dir.register(watchService, StandardWatchEventKinds.ENTRY_MODIFY);

        try {
            while (dirty) {
                WatchKey key = watchService.take();
                for (WatchEvent<?> watchEvent : key.pollEvents()) {
                    //getting the type of the event
                    WatchEvent.Kind<?> kind = watchEvent.kind();
                    //getting the file name from the WatchEvent object
                    Path eventFile = (Path) watchEvent.context();
                    if (Files.isSameFile(myFile, eventFile)) dirty = false; //kind.equals(StandardWatchEventKinds.ENTRY_MODIFY) &&
                    System.out.println("Filename --> " + eventFile + " Event type --> " + kind.name());
                }
                //reset the current event and wait for other events
                if (!key.reset()) break;
            }
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        */
    }

    public int size() {
        return pageBuffer.size();
    }

    /**
     * Pins a page.
     * Such pages can not be swapped out of the buffer. The pinCount
     * in DBBufferStoredElement is increased by 1.
     *
     * @param pageID   the page ID of the page to be pinned
     */
    public void pin(int pageID) {
        CBufferElement stored = find(pageID);
        if (stored != null) {
            stored.pinCount++;
        }
    }

    /**
     * Unpins a page.
     * The pinCount in DBBufferStoredElement is decreased by 1.
     *
     * @param pageID   the page ID of the page to be unpinned
     */
    public void unpin(int pageID) {
        CBufferElement stored = find(pageID);
        if ( stored != null && stored.pinCount > 0 ) {
            stored.pinCount--;
        }
    }

    public void close() throws IOException {
        flush();
        if (file != null) {
            file.close();
            file = null;
        }
    }

    /* *********************** *
     *    abstract methods     *
     * *********************** */

    protected abstract void savePage(CBufferElement page) throws IOException;

}
