package database;

import java.io.Serializable;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Types;
import java.util.Properties;

/**
 * Connects to a database.
 * <p/>
 * This class is a modification of the work done by:
 *
 * @author Len Trigg (trigg@cs.waikato.ac.nz)
 * @author Stefan Mutter (mutter@cs.waikato.ac.nz)
 * @version $Revision: 1.2 $
 *          <p/>
 *          Modified: January 2006
 *          By: Avare Stewart
 *          Modified: January 2011
 *          By: Mikalai Tsytsarau
 */
public class DB implements Serializable {

    /**
     * The name of the properties file
     */
    private static String CONFIG_PATH = "./config/";
    protected static String PROPERTY_FILE = CONFIG_PATH + "db.properties";

    /** Holds the jdbc drivers to be used (only to stop them being gc'ed) */
    //protected static Vector DRIVERS = new Vector();


    /** Properties associated with the database connection */
    protected static Properties PROPERTIES;

    /**
     * Database URL
     */
    protected String m_DatabaseURL = null;

    /**
     * Database URL
     */
    protected String m_driver = null;

    /**
     * The prepared statement used for database queries.
     * Sometimes it is more convenient or more efficient to use a PreparedStatement object for sending SQL statements to the database. This special type of statement is derived from the more general class, Statement, that you already know.
     * When to Use a PreparedStatement Object: If you want to execute a Statement object many times, it will normally reduce execution time to use a PreparedStatement object instead.
     * The main feature of a PreparedStatement object is that, unlike a Statement object, it is given an SQL statement when it is created. The advantage to this is that in most cases, this SQL statement will be sent to the DBMS right away, where it will be compiled. As a result, the PreparedStatement object contains not just an SQL statement, but an SQL statement that has been precompiled. This means that when the PreparedStatement is executed, the DBMS can just run the PreparedStatement 's SQL statement without having to compile it first.
     * Although PreparedStatement objects can be used for SQL statements with no parameters, you will probably use them most often for SQL statements that take parameters. The advantage of using SQL statements that take parameters is that you can use the same statement and supply it with different values each time you execute it. You will see an example of this in the following sections.
     */
    protected PreparedStatement m_PreparedStatement;

    /**
     * The prepared statement used for database queries.
     */
    protected static Statement m_Statement = null;

    /**
     * The database connection
     */
    protected static Connection m_Connection;

    /**
     * True if debugging output should be printed
     */
    protected boolean m_Debug = false;

    /* For databases where Tables and Columns are created in upper case */
    protected boolean m_checkForUpperCaseNames = false;

    /* setAutoCommit on the database? */
    protected boolean m_setAutoCommit = true;

    /* create index on the database? */
    protected boolean m_createIndex = false;

    /**
     * Database username
     */
    protected String m_userName = null;

    /**
     * Database Password
     */
    protected String m_password = null;


    /* Type mapping used for reading experiment results */
    public static final int STRING = 0;
    public static final int BOOL = 1;
    public static final int DOUBLE = 2;
    public static final int BYTE = 3;
    public static final int SHORT = 4;
    public static final int INTEGER = 5;
    public static final int LONG = 6;
    public static final int FLOAT = 7;
    public static final int DATE = 8;

    public DB(Properties props) {

        if (PROPERTIES == null || PROPERTIES.isEmpty()) {
            //loading up the properties file
            PROPERTIES = props;
        }

        //reading the properties for the MySQL Connection
        String hostname = PROPERTIES.getProperty("hostname");
        String port = PROPERTIES.getProperty("port");
        String user = PROPERTIES.getProperty("user");
        String password = PROPERTIES.getProperty("password");
        String database = PROPERTIES.getProperty("database");
        String driver = PROPERTIES.getProperty("driver");
        String jdbc = PROPERTIES.getProperty("jdbc");
        String url = jdbc + "//" + hostname + ":" + port + "/" + database;

        init(user, password, url, driver);
    }

    public DB(String user, String password, String url, String driver) {
        init(user, password, url, driver);
    }

    public void init(String user, String password, String url, String driver) {
        DriverManager.setLoginTimeout(60);
        this.m_userName = user;
        this.m_password = password;
        this.m_DatabaseURL = url;
        this.m_driver = driver;
    }

    /**
     * Set the database username
     *
     * @param username Username for Database.
     */
    public void setUsername(String username) {
        m_userName = username;
    }


    /**
     * Get the DatabaseStatement.
     *
     * @return DatabaseStatement.
     */
    public Statement getDatabaseStatement() {

        return m_Statement;
    }


    /**
     * Get the database username
     *
     * @return Database username
     */
    public String getUsername() {
        return m_userName;
    }

    /**
     * Set the database password
     *
     * @param password Password for Database.
     */
    public void setPassword(String password) {
        m_password = password;
    }

    /**
     * Get the database password
     *
     * @return Password for Database.
     */
    public String getPassword() {
        return m_password;
    }

    /**
     * Get the value of DatabaseURL.
     *
     * @return Value of DatabaseURL.
     */
    public String getDatabaseURL() {

        return m_DatabaseURL;
    }

    /**
     * Set the value of DatabaseURL.
     *
     * @param newDatabaseURL Value to assign to DatabaseURL.
     */
    public void setDatabaseURL(String newDatabaseURL) {

        m_DatabaseURL = newDatabaseURL;
    }

    /**
     * Check if the property checkUpperCaseNames in the DatabaseUtils file is set to true or false.
     *
     * @return true if the property checkUpperCaseNames in the DatabaseUtils file is set to true, false otherwise.
     */
    public boolean getUpperCase() {
        return m_checkForUpperCaseNames;
    }


    /**
     * Converts an array of objects to a string by inserting a space
     * between each element. Null elements are printed as ?
     *
     * @param array the array of objects
     * @return a value of type 'String'
     */
    public static String arrayToString(Object[] array) {

        String result = "";
        if (array == null) {
            result = "<null>";
        } else {
            for (int i = 0; i < array.length; i++) {
                if (array[i] == null) {
                    result += " ?";
                } else {
                    result += " " + array[i];
                }
            }
        }
        return result;
    }

    /**
     * Returns the name associated with a SQL type.
     *
     * @param type the SQL type
     * @return the name of the type
     */
    public static String typeName(int type) {

        switch (type) {
            case Types.BIGINT:
                return "BIGINT ";
            case Types.BINARY:
                return "BINARY";
            case Types.BIT:
                return "BIT";
            case Types.CHAR:
                return "CHAR";
            case Types.DATE:
                return "DATE";
            case Types.DECIMAL:
                return "DECIMAL";
            case Types.DOUBLE:
                return "DOUBLE";
            case Types.FLOAT:
                return "FLOAT";
            case Types.INTEGER:
                return "INTEGER";
            case Types.LONGVARBINARY:
                return "LONGVARBINARY";
            case Types.LONGVARCHAR:
                return "LONGVARCHAR";
            case Types.NULL:
                return "NULL";
            case Types.NUMERIC:
                return "NUMERIC";
            case Types.OTHER:
                return "OTHER";
            case Types.REAL:
                return "REAL";
            case Types.SMALLINT:
                return "SMALLINT";
            case Types.TIME:
                return "TIME";
            case Types.TIMESTAMP:
                return "TIMESTAMP";
            case Types.TINYINT:
                return "TINYINT";
            case Types.VARBINARY:
                return "VARBINARY";
            case Types.VARCHAR:
                return "VARCHAR";
            default:
                return "Unknown";
        }
    }

/*
 * Closes the connection and opens it again
 */

    public void reconnect() throws Exception {
        disconnectFromDatabase();
        connectToDatabase();
    }

    /**
     * Precondition the user name, password and url are set Opens a connection to the database. If the connection information can assuming the connection information is coming from a properties file
     *
     * @throws Exception if an error occurs
     */
    public void connectToDatabase() throws Exception {
        if (m_Connection == null) {
            assert m_DatabaseURL == null : "the url, password and user name must be set";
            //making the connection
            try {
                // The newInstance() call is a work around for some
                // broken Java implementations
                //Class.forName("com.mysql.jdbc.Driver").newInstance();
                //Class.forName("com.ibm.db2.jcc.DB2Driver").newInstance();
                Class.forName(m_driver).newInstance();
                m_Connection = DriverManager.getConnection(m_DatabaseURL, m_userName, m_password);
            }
            catch (ClassNotFoundException e) {
                System.err.println("Cannot load jdbc Driver: " + e);
            } catch (InstantiationException e) {
                System.err.println("Cannot instantiate Class: " + e);
                e.printStackTrace();
            } catch (IllegalAccessException e) {
                System.err.println("Illegal Access to Class: " + e);
                e.printStackTrace();
            } catch (SQLException e) {
                System.err.println("Error when making connection. Maybe wrong hostname, user or password?: " + e);
            }
        }
        if (m_Connection != null) {
            m_Statement = m_Connection.createStatement();
            System.out.println("\n\n **** Database Connection Established ****");
        } else {
            throw new Exception();
        }
    }

    /**
     * Closes the connection to the database.
     *
     * @throws Exception if an error occurs
     */
    public void disconnectFromDatabase() throws Exception {

        if (m_Debug) {
            System.err.println("Disconnecting from " + m_DatabaseURL);
        }
        if (m_Connection != null) {
            m_Connection.close();
            m_Connection = null;
        }
    }

    /**
     * Returns true if a database connection is active.
     *
     * @return a value of type 'boolean'
     */
    public boolean isConnected() {

        return (m_Connection != null);
    }

    /**
     * Executes a Prepeard SQL query.
     *
     * @param query the SQL query
     * @return true if the query generated results
     * @throws SQLException if an error occurs
     */
    public ResultSet execute(String query) throws SQLException {

        ResultSet rs = null;
        boolean val = m_Statement.execute(query);
        if (val) {
            rs = this.getResultSet();
        }

        return rs;

    }

    /**
     * Gets the results generated by a previous query.
     *
     * @return the result set.
     * @throws SQLException if an error occurs
     */
    public ResultSet getResultSet() throws SQLException {

        return m_Statement.getResultSet();
    }

    public Connection getConnection() {
        return m_Connection;
    }
}

