package main;

/**
 * A class for accessing statistical sentiment values including:
 * average, standard deviation, volume, positive and negative, etc.
 * from 1-st and 2-nd order sentiment moments stored in {@link main.Values}
 *
 * @author Mikalai Tsytsarau;
 */

import utils.Statistics;
import java.util.Comparator;
import java.util.List;
import java.util.Locale;

public class CValues extends Values implements Comparable<CValues>, Cloneable {
    public static float CONTR_THRESHOLD = 0.5f;

    public static final int TYPE_CONTR_SYNC = 2;
    public static final int TYPE_CONTR_ASYNC = 4;
    public static final int TYPE_CONTR_ZERO = 8;
    public static final int TYPE_CONTR_ALL = 14;
    public static final int TYPE_CONTR_HIGH = 6;

    public static final int DAY = 86400;
    public static final int NONE = -1;
    public static final int M1 = 6;
    public static final int M2 = 7;
    public static final int ID = 8;
    public static final int TIME = 5;
    public static final int VOLUME = 0;
    public static final int VOLUME_POS = 12;
    public static final int VOLUME_NEG = 13;
    public static final int WEIGHT = 9;
    public static final int PRIME = 10;
    public static final int DOUBLEPRIME = 11;
    public static final int VARIANCE = 1;
    public static final int AVERAGE = 2;
    public static final int CORRELATION = 3;
    public static final int CONTRADICTION = 4;

    public TimeSeries ts = null;
    public int id = -1;
    public int begin = -1;
    public int end = -1;

    public float c = 0f; // contr
    public float p = 0f; // significance/prob

    public static float alpha = 0.005f;
    public static final float[] weights = new float[]{1.0049453f, 1.0359724f, 1.2384058f, 2.0f, 2.7615943f, 2.9640276f, 2.9950547f, 2.9993293f, 2.9999092f, 2.9999876f};
    public int type = TYPE_CONTR_ZERO;

    public CValues() {

    }

    public CValues(int id, float c) {
        this.id = id;
        this.c = c;
    }

    public CValues(int id, int begin, float s) {
        this(id, begin, begin);
        super.update(s, 1);
    }

    public CValues(int id, int begin, int end, Values values) {
        this(id, begin, end);
        super.update(values, 1);
    }

    public CValues(int id, int begin, int end, float n, float m1, float m2) {
        this(id, begin, end);
        this.n = n;
        this.m1 = m1;
        this.m2 = m2;
    }
    
    public CValues(int id, int begin, int end, List<Float> sentiments) {
        this(id, begin, end);
        for (Float s : sentiments) {
            m1 += s;
            m2 += s * s;
        }
        n += sentiments.size();
    }

    public CValues(int id, int begin, int end) {
        this.id = id;
        this.begin = begin;
        this.end = end;
    }

    public CValues(List<? extends Values> toAggregate) {
        for (Values v : toAggregate) {
            update(v);
        }
    }

    public static Float getFloat(Object o) {
        if (o == null) return null;
        if (CValues.class.isInstance(o)) return new Float(((CValues)o).c);
        if (o.getClass() == Double.class) return new Float((Double)o);
        if (o.getClass() == Float.class) return new Float((Float)o);
        if (o.getClass() == Byte.class) return new Float(((Byte)o) / 127f);
        return null;
    }

    public void update(CValues values) {
        super.update(values);
        this.p += values.p;
    }

    public void reset() {
        n = 0f;
        m1 = 0f;
        m2 = 0f;
        c = 0f;
        p = 0f;
    }

    public float getVal(int type) {
        switch (type) {
            case M1: return m1;
            case M2: return m2;
            case TIME: return time();
            case NONE: return 1;
            case VOLUME: return n;
            case VOLUME_POS: return getPos();
            case VOLUME_NEG: return getNeg();
            case WEIGHT: return getW();
            case VARIANCE: return getVar();
            case AVERAGE: return getAvg();
            case CONTRADICTION: return getContr();
        }
        return 0;
    }

    public float getPos() {
        float avg = 0;
        float mean = getAvg(); float stdev = getStdev();
        return (float) Statistics.normalProbability((mean - avg) / stdev) * getWeight(n);
    }

    public float getNeg() {
        return getWeight(n) - getPos();
    }

    public void setVal(int type, float v) {
        switch (type) {
            case M1: m1 = v; break;
            case M2: m2 = v; break;
            case VOLUME: n = v; break;
            //case VARIANCE: vt = v; break;
            //case AVERAGE: st = v; break;
            case CONTRADICTION: c = v; break;
            case CORRELATION: c = v; break;
        }
    }

    public float getContr() {
        if (n <= 0) return Float.NaN;
        float avg = 0;
        float mu = m1 / n - avg; float m = mu * mu;
        c = alpha * (m2 / n - m) / (alpha + m);
        c *= getW();
        return c;
    }

    public float getW() {
        return getWeight(n);
    }

    public static float getWeight(float nn) {
        int i = Math.max(0, Math.round(nn / 3.33f)); //middle value is for n = 10
        return (i < weights.length) ? 0.5f * weights[i] - 0.5f : 1f; // weight
    }

    public CValues clone() {
        CValues v = new CValues(id, begin, end, n, m1, m2);
        v.c = c; v.p = p; v.type = type;
        v.np = np; v.nn = nn; v.no = no;
        return v;
    }

    public long timestamp() {
        return (begin * 500L + end * 500L);
    }

    public int time() {
        return (begin / 2 + end / 2);
    }

    public long interval() {
        return (long) (end - begin);
    }

    public static boolean intersect(CValues ri, CValues rj) {
        return (Math.signum(ri.end - rj.begin) * Math.signum(rj.end - ri.begin) > 0); // without signum * causes overflows
    }

    public boolean intersects(CValues v) {
        return intersect(this, v);
    }

    public static int distance(CValues ri, CValues rj) {
        return Math.abs((ri.begin - rj.begin) + (ri.end - rj.end)) / 2;
    }

    public static int diff(CValues ri, CValues rj) {
        return ((ri.begin - rj.begin) + (ri.end - rj.end)) / 2;
    }


    public static int contradictionType(double c) {
        return contradictionType(c, CONTR_THRESHOLD);
    }

    public static int contradictionType(double c, double threshold) {
        if (c > threshold) {
            return TYPE_CONTR_SYNC;
        }else if (c < -threshold) {
            return TYPE_CONTR_ASYNC;
        }
        return TYPE_CONTR_ZERO;
    }

    public String toString() {
        java.text.SimpleDateFormat stub = new java.text.SimpleDateFormat("dd.MM.yyyy");
        return stub.format(new java.util.Date(begin * 1000L)) + "-" + stub.format(new java.util.Date(end * 1000L))
                + " id:" + id + " c:" + c;
    }

    public String getString() {
        return String.format(Locale.US, "%d\t%d\t%.3f\t%.3f\t%.3f\t%.3f", begin, end, n, getAvg(), getVar(), getContr());
    }

    public static Comparator<CValues> LowFirstSortTime = new Comparator<CValues>() {
        @Override
        public int compare(CValues v1, CValues v2) {
            return Math.round(Math.signum(v1.timestamp() - v2.timestamp())); // lowest first
        }
    };

    public static Comparator<CValues> LowFirstSortC = new Comparator<CValues>() {
        @Override
        public int compare(CValues c1, CValues c2) {
            return Math.round(Math.signum(Math.abs(c1.c) - Math.abs(c2.c))); // lowest first
        }
    };

    public static Comparator<CValues> HighFirstSortC = new Comparator<CValues>() {
        @Override
        public int compare(CValues c1, CValues c2) {
            return Math.round(Math.signum(Math.abs(c2.c) - Math.abs(c1.c))); // highest first
        }
    };

    public static Comparator<CValues> HighFirstSortCP = new Comparator<CValues>() {
        @Override
        public int compare(CValues c1, CValues c2) {
            return (int) Math.round(Math.signum(Math.abs(c2.c) + Math.log1p(c2.p) - Math.abs(c1.c) - Math.log1p(c1.p))); // highest first
        }
    };

    public static Comparator<CValues> LowFirstSortP = new Comparator<CValues>() {
        @Override
        public int compare(CValues c1, CValues c2) {
            return Math.round(Math.signum(Math.abs(c1.p) - Math.abs(c2.p))); // lowest first
        }
    };

    public int compareTo(CValues o) {
        return (int)Math.signum(o.c - this.c);  // highest first
    }

    public int hashCode() {
        return id;
    }

    public boolean equals(java.lang.Object o) {
        return o.hashCode() == this.hashCode();
    }
}
