package main;

/**
 * A series that holds statistical sentiment values {@link main.CValues}
 *
 * @author Mikalai Tsytsarau;
 */

import java.io.*;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.*;

public class TimeSeries extends CValues implements Cloneable, List<CValues> {

    public int topicId = -1;
    public int granularity;
    public int window = DAY;
    public CValues last = null;
    ArrayList<CValues> ts = null;

    public TimeSeries() {
        ts = new ArrayList<CValues>();
    }

    public TimeSeries(int topicId, int id, int begin, int end, int window) {
        this.topicId = topicId;
        this.id = id;
        this.begin = begin;
        this.end = end;
        this.window = window;
        this.ts = new ArrayList<CValues>();
    }

    public String toString() {
        DateFormat df = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss");
        return "Topic " + topicId + " [" + df.format(new Date(begin * 1000L)) + " - " + df.format(new Date(end * 1000L)) + "]";
    }

    public float[] stats(int type) {
        return stats(type, true);
    }

    public float[] stats(int type, boolean weighted) {
        Values stats = new Values();
        float max = Float.MIN_VALUE; float min = Float.MAX_VALUE;
        for (int i = 0; i < size(); i++) {
            CValues v = get(i);
            float len = weighted ? (float) v.interval() / this.interval() : 1;
            float val = v.getVal(type);
            stats.update(val, len);
            if (max < val) max = val;
            if (min > val) min = val;
        }
        return new float[] {min, max, fixNaN(stats.getAvg()), fixNaN(stats.getStdev())};
    }

    public CValues getLast() {
        last = ts.isEmpty() ? null : ts.get(ts.size() - 1);
        return last;
    }

    public void reset() {
        super.reset();
        clear();
    }



    public void outputSeries(String file) {
        outputSeries(file, null);
    }

    public void outputSeries(String file, String title) {
        try {
            BufferedWriter output = new BufferedWriter(new FileWriter(file, true));
            if (title != null) {
                output.write(title + " | ");
                output.write("begin\tend\tn\tavg\tvar\tc"); output.newLine();
            }
            for (CValues v : ts) {
                if (v.n > 0) {
                    output.write(v.getString()); output.newLine();
                }
            }
            output.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public void sort() {
        Collections.sort(ts, CValues.LowFirstSortTime);
        getLast();
    }

    public TimeSeries clone() {
        return clone(true);
    }

    public TimeSeries clone(boolean all) {
        return clone(all, begin, end);
    }

    public TimeSeries clone(boolean all, double start, double finish) {
        TimeSeries clone = new TimeSeries(topicId, id, begin, end, window);
        clone.granularity = this.granularity;
        if (all) {
            clone.update(this, 1);
            clone.c = c;
            clone.p = p;
            for (CValues v : ts) {
                int t1 = v.time() + window;
                int t2 = v.time() - window;
                if (t1 > start && t2 < finish)
                    clone.add(v.clone());

            }
        }
        return clone;
    }

    public CValues getInterval(int time) {
        for (CValues v : ts) {
            if (v.begin <= time && v.end > time) return v;
        }
        return null;
    }

    public CValues getNearestInterval(int time) {
        CValues interval = null;
        for (CValues v : ts) {
            if (v.begin <= time) {
                interval = v;
            } else {
                break;
            }
        }
        return interval;
    }

    public int hashCode() {
        return (id << 16) | topicId;
    }

    /*
    public boolean equals(Object o) {
        //return (o instanceof TimeSeries) ? (this.hashCode() == o.hashCode()) : false;
        return (this.hashCode() == o.hashCode());
    }
    */

    @Override
    public int size() {
        return ts.size();
    }

    @Override
    public boolean isEmpty() {
        return ts.isEmpty();
    }

    @Override
    public boolean contains(Object o) {
        return ts.contains(o);
    }

    @Override
    public void clear() {
        last = null;
        ts.clear();
    }

    @Override
    public boolean add(CValues values) {
        return add(values, false);
    }

    public boolean add(CValues values, boolean update) {
        if (values != null) {
            if (update) update(fixNaN(values.getAvg()), 1);
            values.ts = this;
            last = values;
            return ts.add(values);
        }
        return false;
    }

    @Override
    public boolean remove(Object o) {
        boolean result = ts.remove(o);
        getLast();
        return result;
    }

    public CValues remove(int index) {
        CValues result = ts.remove(index);
        getLast();
        return result;
    }

    @Override
    public int indexOf(Object o) {
        return ts.indexOf(o);
    }

    @Override
    public int lastIndexOf(Object o) {
        return ts.lastIndexOf(o);
    }

    @Override
    public ListIterator<CValues> listIterator() {
        return ts.listIterator();
    }

    @Override
    public ListIterator<CValues> listIterator(int index) {
        return ts.listIterator(index);
    }

    @Override
    public List<CValues> subList(int fromIndex, int toIndex) {
        return ts.subList(fromIndex, toIndex);
    }

    @Override
    public boolean containsAll(Collection<?> c) {
        return ts.containsAll(c);
    }

    @Override
    public boolean addAll(Collection<? extends CValues> c) {
        for (CValues v : c) { v.ts = this; }
        boolean result = ts.addAll(c);
        getLast();
        return result;

    }

    @Override
    public boolean addAll(int index, Collection<? extends CValues> c) {
        for (CValues v : c) { v.ts = this; }
        boolean result = ts.addAll(index, c);
        getLast();
        return result;
    }

    @Override
    public boolean removeAll(Collection<?> c) {
        boolean result = ts.removeAll(c);
        getLast();
        return result;
    }

    @Override
    public boolean retainAll(Collection<?> c) {
        return ts.retainAll(c);
    }

    public CValues get(int index) {
        return ts.get(index);
    }

    @Override
    public CValues set(int index, CValues element) {
        element.ts = this;
        CValues result =  ts.set(index, element);
        getLast();
        return result;
    }

    @Override
    public void add(int index, CValues element) {
        element.ts = this;
        ts.add(index, element);
        getLast();
    }

    @Override
    public Iterator<CValues> iterator() {
        return ts.iterator();
    }

    @Override
    public Object[] toArray() {
        return ts.toArray();
    }

    @Override
    public <T> T[] toArray(T[] a) {
        return ts.toArray(a);
    }

    public static String getName(int type) {
        switch (type) {
            case TimeSeries.AVERAGE:
                return "Sentiment";
            case TimeSeries.CONTRADICTION:
                return "Contradiction";
            case TimeSeries.CORRELATION:
                return "Correlation";
            case TimeSeries.VOLUME:
                return "Volume";
            case TimeSeries.VOLUME_POS:
                return "PosVolume";
            case TimeSeries.VOLUME_NEG:
                return "NegVolume";
            case TimeSeries.VARIANCE:
                return "Variance";
            case TimeSeries.PRIME:
                return "Prime";
            case TimeSeries.DOUBLEPRIME:
                return "DoublePrime";
        }
        return "Series";
    }

    public static float getStdev(float[] ts) {
        CValues cv = new CValues();
        for (float v : ts) {
            cv.update(v);
        }
        return cv.getStdev();
    }

    public static float fixNaN(double value) {
        return (Double.isNaN(value) || Double.isInfinite(value)) ? 0 : (float) value;
    }
}
