package ui;

/**
 * A class for generating and testing {@link ctree.CIndex} and database implementations of a CTree.
 *
 * @author Mikalai Tsytsarau;
 */

import ctree.*;
import database.DatabaseAccess;
import main.CValues;
import main.TimeSeries;
import main.Values;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.DateAxis;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.NumberTickUnit;
import org.jfree.chart.event.ChartProgressEvent;
import org.jfree.chart.event.ChartProgressListener;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.DefaultXYItemRenderer;
import org.jfree.chart.renderer.xy.StandardXYBarPainter;
import org.jfree.chart.renderer.xy.XYBarRenderer;
import org.jfree.chart.renderer.xy.XYDotRenderer;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import test.TestCIndex;
import utils.CubicSpline;

import javax.swing.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.*;
import java.lang.ref.WeakReference;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.DecimalFormat;
import java.util.*;
import java.util.concurrent.CountDownLatch;

public class EvaluationManager extends JFrame implements ActionListener, WindowListener, ChartProgressListener {
    
    public JProgressBar progressBar;
    public JProgressBar progressBarSub;
    public JPanel panelStats;
    public JPanel panelTest;
    public JPanel panelMain;
    public ChartPanel chartPanelStats;
    public ChartPanel chartPanelTest;
    public ChartPanel chartPanelMain;
    public XYPlot plotStats = new XYPlot(); // test
    public XYPlot plotTest = new XYPlot(); // test
    public JFreeChart chartStats;
    public JFreeChart chartTest;
    public CountDownLatch chartLatchTest = null;

    public XYSeries seriesPerformance = new XYSeries("Index performance");
    public XYSeries seriesMaster = new XYSeries("Master Sentiments");
    public XYSeries seriesMasterZeros = new XYSeries("Zero Points");

    private JMenu menu;
    private JMenuItem mnGenerateIndex;
    private JMenuItem mnReplicateTopics;
    private JMenuItem mnExportToDB;
    private JMenuItem mnSmokeTest;
    private JMenuItem mnGenerateUpdates;
    private JMenuItem mnTestUpdatesDB;
    private JMenuItem mnTestUpdatesCTree;
    private JMenuItem mnTestTopKCTree;
    private JCheckBoxMenuItem cbVisualizeEvaluation;
    private static String OUTPUT_PATH = "./out/";
    private static String CONFIG_PATH = "./config/";
    public boolean visualize = false;

    TestCIndex test;
    CModel model;
    DBModel dbModel;
 //   BufferedWriter debugOut;
    int numTopics;
    private long startTime;
    private long stopTime;
    private java.util.List<CValues> topics = null;

    public static void main(String[] args) throws IOException {
        Properties prop = new Properties();
        try {
            FileInputStream input = new FileInputStream(CONFIG_PATH + "db.properties");
            prop.load(input);
        } catch (FileNotFoundException e) {
            System.err.println("Can't find " + CONFIG_PATH + "db.properties");
            return;
        } catch (IOException e) {
            System.err.println("I/O failed.");
            return;
        }

        try {
            DatabaseAccess.init(prop);
        } catch (Exception ex) {
            System.out.println("Error creating jdbc connection. Check Settings.");
        }
        EvaluationManager em = new EvaluationManager();
        em.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        em.setSize(1280, 640);
        em.setVisible(true);
    }

    public EvaluationManager() {
        this.model = new CModel();
        try {
            model.InitializeIndex(null);
        } catch (IOException e) {
            e.printStackTrace();
        }

        test = new TestCIndex(this, this.model);
        menu = new JMenu("Evaluation");
        JMenuBar mb = new JMenuBar();
        mb.add(menu);
        setJMenuBar(mb);
        setSize(960, 640);

        addMenus();
        pack();
        InitializeInterface();
        addWindowListener(this);
        setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);

        dbModel = new DBModel();
        if (!DatabaseAccess.initialized()) {
            mnSmokeTest.setEnabled(false);
            mnExportToDB.setEnabled(false);
            mnTestUpdatesDB.setEnabled(false);
        }
        chartTest.addProgressListener(this);
        
        setContentPane(panelMain);
    }

    public void addMenus() {
        addMenuItem(menu, mnGenerateIndex = new JMenuItem("Generate Index"));
        addMenuItem(menu, mnReplicateTopics = new JMenuItem("Replicate Topics"));
        addMenuItem(menu, mnExportToDB = new JMenuItem("Export to DB"));
        addMenuItem(menu, cbVisualizeEvaluation = new JCheckBoxMenuItem("Visualize Evaluation"));
        menu.addSeparator();
        addMenuItem(menu, mnSmokeTest = new JMenuItem("Smoke Test"));
        addMenuItem(menu, mnGenerateUpdates = new JMenuItem("Generate Updates"));
        addMenuItem(menu, mnTestUpdatesDB = new JMenuItem("Test Updates DB"));
        addMenuItem(menu, mnTestUpdatesCTree = new JMenuItem("Test Updates CTree"));
        addMenuItem(menu, mnTestTopKCTree = new JMenuItem("Test TopK Topics"));
    }
    
    private void addMenuItem(JMenu menu, JMenuItem item) {
        menu.add(item);
        item.addActionListener(this);
    }
    
    private void InitializeInterface() {
        XYSeriesCollection datasetPerformance = new XYSeriesCollection();
        XYSeriesCollection datasetMaster = new XYSeriesCollection();
        XYSeriesCollection datasetMasterZeros = new XYSeriesCollection();

        datasetPerformance.addSeries(seriesPerformance); // index performance
        datasetMaster.addSeries(seriesMaster);
        datasetMasterZeros.addSeries(seriesMasterZeros);

        final NumberAxis rangeAxis5 = new NumberAxis(); //"Sentiments"
        rangeAxis5.setNumberFormatOverride(new DecimalFormat("#.#"));
        rangeAxis5.setTickUnit(new NumberTickUnit(1));
        rangeAxis5.setMinorTickCount(5);
        rangeAxis5.setMinorTickMarksVisible(true);
        rangeAxis5.setRange(-1, 1);

        final DateAxis dateAxis2 = new DateAxis("", TimeZone.getTimeZone("PST"), Locale.US);
        dateAxis2.setDateFormatOverride(CModel.formatMonth);

        final XYBarRenderer renderer1 = new XYBarRenderer();
        renderer1.setBarPainter(new StandardXYBarPainter());
        renderer1.setSeriesPaint(0, new Color(0, 0, 200, 40));
        renderer1.setDrawBarOutline(false);
        renderer1.setShadowVisible(false);

        final DefaultXYItemRenderer renderer23 = new DefaultXYItemRenderer();
        renderer23.setSeriesPaint(0, Color.BLUE);
        renderer23.setBaseShapesVisible(false);

        final XYDotRenderer renderer4 = new XYDotRenderer();
        renderer4.setDotHeight(8);
        renderer4.setDotWidth(8);
        renderer4.setBasePaint(Color.MAGENTA);

        //final LogAxis logAxis0 = new LogAxis();
        //logAxis0.setTickUnit(new NumberTickUnit(1));
        final NumberAxis rangeAxis1 = new NumberAxis();
        rangeAxis1.setRange(0, 1e6);

        plotStats.setDataset(0, datasetPerformance);
        plotStats.setRenderer(0, renderer23);
        plotStats.setRangeAxis(rangeAxis1); //
        plotStats.setDomainAxis(new NumberAxis());

        plotTest.setDataset(0, datasetMaster);
        plotTest.setDataset(1, datasetMasterZeros);
        plotTest.setRenderer(0, renderer23);
        plotTest.setRenderer(1, renderer4);
        plotTest.setRangeAxis(rangeAxis5);
        plotTest.setDomainAxis(dateAxis2);


        // add the subplots...
        chartStats = new JFreeChart(plotStats);
        chartStats.removeLegend();
        chartStats.setBackgroundPaint(Color.white);
        chartTest = new JFreeChart(plotTest);
        chartTest.removeLegend();
        chartTest.setBackgroundPaint(Color.white);
        
        panelStats.setLayout(new GridLayout());
        panelStats.add(chartPanelStats = new ChartPanel(chartStats));
        panelTest.setLayout(new GridLayout());
        panelTest.add(chartPanelTest = new ChartPanel(chartTest));
    }

    public void actionPerformed(ActionEvent event) {
        Object src = event.getSource();
        if (src == mnSmokeTest) test.smokeTest();
        if (src == mnGenerateIndex) generateIndexAsync(1000);
        if (src == mnReplicateTopics) replicateTopicsAsync(10000);
        if (src == mnExportToDB) copyIndexToDbAsync();
        if (src == mnGenerateUpdates) generateUpdates();
        if (src == mnTestUpdatesDB) testUpdates(true);
        if (src == mnTestUpdatesCTree) testUpdates(false);
        if (src == mnTestTopKCTree) testTopKAsync(2, false);
        if (src == cbVisualizeEvaluation) visualize = cbVisualizeEvaluation.isSelected();
    }

    protected void finalize () {
        model.FreeResources();
        try {
            super.finalize();
        } catch (Throwable throwable) {
            throwable.printStackTrace();
        }
    }


    public void testTimeIntervalAsync(final int queryType, final boolean useAdaptiveThreshold, final boolean useDatabase) {
        new Thread(new Runnable() {
            public void run() {
                testTimeInterval(queryType, useAdaptiveThreshold, useDatabase);
            }
        }).start();
    }
    public void testTimeInterval(final int queryType, final boolean useAdaptiveThreshold, final boolean useDatabase) {
        String fileName = OUTPUT_PATH + "test_time_" + queryType + (useDatabase ? "_dbms_" : "_tree_") + (useAdaptiveThreshold ? "adapt" : "fixed") + ".dat";
        System.out.println(fileName);

        seriesPerformance.clear();

        int m = 5;
        if (useDatabase && queryType == 2) m = 1;

        Model model = (useDatabase ? dbModel : this.model);

        int n = 10;
        int step = (this.model.indexEnd - this.model.indexBegin)/n;
        long[] times = new long[n];

        float[] thresholds = useAdaptiveThreshold ?
                new float[]{23.0f, 0.95f, 1.073f, 1.02f} :
                new float[]{78.05f, 50f, 85.05f, 81.05f};
        int[] topic_ids = new int[]{7, 217, 689, 410};

        progressBar.setMaximum(m * n);
        progressBar.setMinimum(0);
        progressBar.setStringPainted(true);

        for (int j = 0; j < m; j++) {
            for (int i = 0; i < n; i++) {
                int t = this.model.indexBegin + step * (i + 1);
                gc();
                startTime = System.currentTimeMillis(); // start timing
                //startTime = mx.getCurrentThreadCpuTime();
                for (int k = 0; k < 4; k++) {
                    if (queryType == 1) {
                        topics = model.getContradictions(topic_ids[k], this.model.indexBegin, t, k, thresholds[k]*0.3f, useAdaptiveThreshold);
                    } else {
                        topics = model.getContradictingTopics(this.model.indexBegin, t, k, thresholds[k], useAdaptiveThreshold, 0);
                    }
                }
                stopTime = System.currentTimeMillis(); // stop timing
                //stopTime = mx.getCurrentThreadCpuTime();
                //if (j > 0) {
                    times[i] += (stopTime - startTime);
                //}
                progressBar.setValue((i + 1) * (j + 1));
                progressBar.invalidate();
            }
        }
        progressBar.setValue(progressBar.getMinimum());
        try {
            BufferedWriter out = new BufferedWriter(new FileWriter(fileName));
            for (int i = 0; i < n; i++) {
                out.write((step * (i + 1)) + " " + ((float) times[i]) / (m )); out.newLine();
                seriesPerformance.add((step * (i + 1)), ((float) times[i]) / (m ));
            }
            out.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public void testGranularityAsync(final int queryType, final boolean useAdaptiveThreshold, final boolean useDatabase, final int limit) {

        new Thread(new Runnable() {
            public void run() {
                testGranularity(queryType, useAdaptiveThreshold, useDatabase, limit);
            }
        }).start();
    }

    public void testGranularity(int queryType, boolean useAdaptiveThreshold, boolean useDatabase, int limit) {
        String fileName = OUTPUT_PATH + "test_gran_" + queryType + (useDatabase ? "_dbms_" : "_tree_") + (useAdaptiveThreshold ? "adapt_limit_" : "fixed_limit_") + limit + ".dat";
        System.out.println(fileName);
        Model model = (useDatabase ? dbModel : this.model);
        seriesPerformance.clear();
        int interval = this.model.indexEnd - this.model.indexBegin;
        int m = (useDatabase && queryType == 2) ? 2 : 11;
        if (limit <= 0 || limit > 4)  limit = 4;
        int n = 4;
        long[] times = new long[n];

        float[] thresholds = useAdaptiveThreshold ?
                new float[]{23.0f, 0.95f, 1.073f, 1.02f} :
                new float[]{78.05f, 50f, 85.05f, 81.05f};
        int[] multipliers = new int[]{3, 5, 7, 9};
        int[] topic_ids = new int[]{243 * (limit - 1) + 60, 243 * (limit - 1) + 70, 243 * (limit - 1) + 80, 243 * (limit - 1) + 90};
        progressBar.setMaximum(n*m);
        progressBar.setMinimum(0);
        progressBar.setStringPainted(true);

        for (int j = 0; j < m; j++) {
            for (int i = 0; i < n; i++) {
                gc();
                startTime = System.currentTimeMillis(); // start timing
                //startTime = mx.getCurrentThreadCpuTime();
                for (int k = 0; k < 4; k++) {
                    if (queryType == 1) {
                        topics = model.getContradictions(topic_ids[k], this.model.indexBegin, this.model.indexBegin + multipliers[k]*interval/10, i, thresholds[i], useAdaptiveThreshold);
                    } else {
                        topics = model.getContradictingTopics(this.model.indexBegin, this.model.indexBegin + multipliers[k]*interval/10, i, thresholds[i], useAdaptiveThreshold, limit);
                    }
                }
                stopTime = System.currentTimeMillis(); // stop timing
                //stopTime = mx.getCurrentThreadCpuTime();
                if (j > 0) {
                    times[i] += (stopTime - startTime);
                }
                progressBar.setValue((i + 1) * (j + 1));
                progressBar.invalidate();

            }
        }
        progressBar.setValue(progressBar.getMinimum());

        try {
            BufferedWriter out = new BufferedWriter(new FileWriter(fileName));
            for (int i = 0; i < n; i++) {
                out.write(i + " " + ((float) times[i]) / (m - 1)); out.newLine();
                seriesPerformance.add(i, ((float) times[i]) / (m - 1));
            }
            out.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public void testTopKAsync(final int queryType, final boolean useDatabase) {
        new Thread(new Runnable() {
            public void run() {
                testTopK(queryType, useDatabase);
            }
        }).start();
    }

    public void testTopK(int queryType, boolean useDatabase) {
        String fileName = OUTPUT_PATH +
                "test_topk_" + (queryType == 1 ? "contr" : "topics") +
                (useDatabase ? "_dbms" : "_tree") +
                (model.index.sorted ? "_sort" : "_unsort") + ".dat";
        System.out.println(fileName);

        seriesPerformance.clear();
        int interval = model.indexEnd - model.indexBegin;

        Model model = (useDatabase ? dbModel : this.model);

        int num_runs = 1;
        //if (useDatabase && queryType == 2) num_runs = 2;
        int num_points = 10;
        int max_height = this.model.index.height;

        long[] times = new long[num_points];
        int[] multipliers = new int[]{3, 5, 7, 9};
        int[] topic_ids = new int[]{43, 70, 122, 165};
        progressBar.setMaximum(num_runs * num_points * max_height);
        progressBar.setMinimum(0);
        progressBar.setStringPainted(true);

        for (int i = 0; i < num_runs; i++) {
            for (int j = 0; j < num_points; j++) {
                for (int height = 0; height < max_height; height++) {
                    gc();
                    startTime = System.currentTimeMillis(); // start timing
                    //startTime = mx.getCurrentThreadCpuTime();
                    for (int k = 0; k < 4; k++) {
                        if (queryType == 1) {
                            topics = model.getContradictionsTopK(topic_ids[k], this.model.indexBegin, this.model.indexBegin + multipliers[k]*interval/10, height, 23 + 244 * j);
                        } else {
                            topics = model.getContradictingTopicsTopK(this.model.indexBegin, this.model.indexBegin + multipliers[k]*interval/10, height, 23 + 244 * j);
                        }
                    }
                    stopTime = System.currentTimeMillis(); // stop timing
                    //stopTime = mx.getCurrentThreadCpuTime();
                    if (i > 0 || !useDatabase) { // skip the first run
                        times[j] += (stopTime - startTime);
                    }
                    progressBar.setValue(progressBar.getValue() + 1);
                    progressBar.invalidate();

                }
            }
        }
        progressBar.setValue(progressBar.getMinimum());

        if (useDatabase) {
            num_runs--;
        }

        try {
            BufferedWriter out = new BufferedWriter(new FileWriter(fileName));
            for (int j = 0; j < num_points; j++) {
                out.write(23 + 244 * j + "\t" + ((float) times[j]) / num_runs); out.newLine();
                seriesPerformance.add(23 + 244 * j, ((float) times[j]) / num_runs);
            }
            out.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public void generateIndexAsync(int _numTopics) {
        numTopics = _numTopics;

        new Thread(new Runnable() {
            public void run() {
                try {
                    generateIndex();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }).start();
    }

    public void generateIndex() throws IOException {
        progressBar.setMaximum(model.indexEnd);
        progressBar.setMinimum(model.indexBegin);
        progressBar.setStringPainted(true);
        seriesPerformance.clear();
        seriesMaster.clear();
        invalidate();

        float[][] series = new float[numTopics][]; int length = Integer.MAX_VALUE;
        for (int i = 0; i < numTopics; i++) {
            double rate = (Math.random() + 0.5) / 5;
            series[i] = generateTimeSeries(model.indexBegin, model.indexEnd, rate);
            length = Math.min(length, series[i].length);
        }

        seriesPerformance.add(0, null); seriesPerformance.add(length, null);
        seriesMaster.add(model.indexBegin*1000L, null); seriesMaster.add(model.indexEnd*1000L, null);

        int t = model.indexBegin;
        for (int j = 0; j < length; j++) {
            long time_begin = System.currentTimeMillis();
            for (int i = 0; i < numTopics; i++) {
                float s = series[i][j];
                float var_s = s*s + (0.05f + 0.05f*(float)Math.random())*(1-Math.abs(s));//no variance can be if average = +/-1
                model.index.insert(i, t, new Values(100, 100 * s, 100 * var_s), 0);
            }
            seriesMaster.add(t*1000L, series[0][j]);
            t += CModel.DAY;
            model.index.flush(); gc();
            long time_end = System.currentTimeMillis();
            progressBar.setValue(t);
            progressBar.invalidate();
            seriesPerformance.add(j, TimeSeries.fixNaN((1000.0 * numTopics) / (time_end - time_begin))); // updates per second

            invalidate();
        }

        progressBar.setValue(progressBar.getMinimum());
    }

    private float[] generateTimeSeries(int begin, int end, double rate) {
        ArrayList<Float> master_s = new ArrayList<Float>();
        ArrayList<Float> master_t = new ArrayList<Float>();
        float s = 0; int t = 0;
        Random rnd_s = new Random();
        Random rnd_t = new Random();
        for (t = begin; t <= end;) {
            master_s.add(s);
            master_t.add((float)t);
            // sample next t from CDF of Poisson distribution +10 days
            int delta_t = (int) (5 - Math.log(1 - rnd_t.nextDouble()) / rate);
            //if (delta_t < 1) delta_t = 1;
            float deviation_s = Math.abs(s) < 0.5 ? 0.2f : 0.1f;
            float delta_s = (float) (deviation_s * (rnd_s.nextDouble() - 0.5 - 0.25 * s));
            if (Math.abs(s + delta_s) > 1) delta_s = Math.signum(s) - s;
            s = s + delta_s;
            t = t + delta_t * CModel.DAY;
        }
        // add to the end to be able to interpolate
        master_s.add(s);
        master_t.add((float)t);

        CubicSpline cs = new CubicSpline(master_t, master_s);
        float[] sentiments = new float[(end - begin + CModel.DAY - 1)/ CModel.DAY];
        t = begin;
        float max_s = 0;
        for (int j = 0; j < sentiments.length; j++) {
            s = (float) cs.interpolate(t);
            //if (Math.abs(s) > 1) s = Math.signum(s);
            if (Math.abs(s) > max_s) max_s = Math.abs(s);
            sentiments[j] = s;
            t += CModel.DAY;
        }
        // normalize
        for (int j = 0; j < sentiments.length; j++) {
            sentiments[j] = sentiments[j] / max_s;
        }
        return sentiments;
    }



    public void replicateTopicsAsync(final int numTopics) {
        this.numTopics = numTopics;
        new Thread(new Runnable() {
            public void run() {
                replicateTopics(numTopics);
            }
        }).start();
    }

    public void replicateTopics(int numTopics) {
        int num = model.index.numPages;
        progressBar.setMaximum(num);
        progressBar.setMinimum(0);
        progressBar.setStringPainted(true);

        seriesPerformance.clear(); seriesPerformance.add(0, null); seriesPerformance.add(num, null);
        invalidate();

        for (int i = 1; i <= num; i++) {
            DBPage page = model.index.readPage(i);
            if (page.pageType == 1) {
                long time_begin = System.currentTimeMillis();
                ((CPage)page).replicateValues(null, 200, numTopics);
                long time_end = System.currentTimeMillis();
                seriesPerformance.add(i, (time_end - time_begin) / 1000.0); // seconds per update
                invalidate();
            }
            progressBar.setValue(i);
        }
        progressBar.setValue(progressBar.getMinimum());
    }

    public void copyIndexToDbAsync() {
        new Thread(new Runnable() {
            public void run() {
                copyIndexToDb();
            }
        }).start();
    }

    public void copyIndexToDb() {
        String query = "";
        try {
            Statement stmt = DatabaseAccess.getDatabaseStatement();

            progressBar.setMaximum(model.index.numPages);
            progressBar.setMinimum(0);
            progressBar.setStringPainted(true);
            DBPage p = null;
            for (int i = 1; i <= model.index.numPages; i++) {
                p = model.index.readPage(i);
                if (p.pageType == 1 || p.pageType == 2) {
                    CPageExtent page = (CPageExtent) p;
                    java.util.List<CValues> topics = page.getPageValues();
                    //System.out.print(page); System.out.println(" " + topics.size());
                    for (int j = 0; j < topics.size(); j++) {
                        CValues topic = topics.get(j);
                        query = "INSERT INTO contradictions VALUES(@topic_id, @timeBegin, @timeEnd, @granularity, @n, @m1, @m2, @contradiction);";
                        query = query.replaceAll("@topic_id", String.valueOf(topic.id));
                        query = query.replaceAll("@timeBegin", String.valueOf(topic.begin));
                        query = query.replaceAll("@timeEnd", String.valueOf(topic.end));
                        query = query.replaceAll("@granularity", String.valueOf(page.height));
                        query = query.replaceAll("@n", String.valueOf(topic.n));
                        query = query.replaceAll("@m1", String.valueOf(topic.m1));
                        query = query.replaceAll("@m2", String.valueOf(topic.m2));
                        query = query.replaceAll("@contradiction", String.valueOf(topic.c));
                        stmt.addBatch(query);
                    }
                    stmt.executeBatch();
                    stmt.clearBatch();
                }
                progressBar.setValue(i);
                progressBar.invalidate();
            }
            progressBar.setValue(progressBar.getMinimum());

        } catch (SQLException ex) {
            System.out.println("SQLException: " + ex.getMessage());
            System.out.println("SQLState: " + ex.getSQLState());
            System.out.println("VendorError: " + ex.getErrorCode());
            System.out.println("Query: " + query);
        }
    }

    public void insert(int topicId, int timestamp, float s) {
        String query = "";
        query = "UPDATE contradictions SET n = n + 1, m1 = m1 + @m1, m2 = m2 + @m2, contradiction = contradict(n, m1, m2) " +
                "WHERE topic_id = @topic_id and (@timestamp between timeBegin and timeEnd);";

        query = query.replaceAll("@topic_id", String.valueOf(topicId));
        query = query.replaceAll("@m1", String.valueOf(s));
        query = query.replaceAll("@m2", String.valueOf(s * s));
        query = query.replaceAll("@timestamp", String.valueOf(timestamp));

        DatabaseAccess.executeUpdate(query);
    }

    public void generateUpdates() {
        FileWriter fstream = null;

        for (int limit = 4; limit <= 40; limit += 4) {
            try {
                fstream = new FileWriter(OUTPUT_PATH + "updates_limit_" + limit + ".dat");
                BufferedWriter out = new BufferedWriter(fstream);
                int counter = 0;
                for (int t = model.indexBegin; t <= model.indexEnd && counter < 1000; t += Math.random() * CModel.DAY * 2) {
                    counter++;
                    int id = (int) (Math.random() * 250 * limit);
                    out.write(id + " " + t + " " + (float) (2 * (Math.random() - 0.5)));
                    out.newLine();
                }
                out.close();

            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    public void testUpdates(boolean useDatabase) {
        String query = "";
        try {
            Statement stmt = DatabaseAccess.getDatabaseStatement();
            int m = 1;
            int n = 10;
            Values[] time_stats = new Values[n];
            for (int i = 0; i < n; i++) {
                time_stats[i] = new Values();
            }
            for (int j = 0; j < m; j++) {
            for (int i = 0; i < n; i++) {
                int limit = (i + 1) * 4;
                FileReader fstream = new FileReader(OUTPUT_PATH + "updates_limit_" + limit + ".dat");
                BufferedReader in = new BufferedReader(fstream);
                int counter = 0;
                String record = null;
                startTime = System.currentTimeMillis(); // start timing
                while ((record = in.readLine()) != null) {
                    String delims = "[ ]+";
                    String[] tokens = record.split(delims);
                    int id = Integer.valueOf(tokens[0]);
                    int timestamp = Integer.valueOf(tokens[1]);
                    float s = Float.valueOf(tokens[2]);
                    if (useDatabase) {
                        query = "UPDATE contradictions SET n = n + 1, m1 = m1 + @m1, m2 = m2 + @m2, contradiction = contradict(n, m1, m2) " +
                                "WHERE topic_id = @topic_id and (@timestamp between timeBegin and timeEnd)";

                        query = query.replaceAll("@topic_id", String.valueOf(id));
                        query = query.replaceAll("@m1", String.valueOf(s));
                        query = query.replaceAll("@m2", String.valueOf(s * s));
                        query = query.replaceAll("@timestamp", String.valueOf(timestamp));

                        stmt.addBatch(query);
                    } else {
                        model.index.insert(id, timestamp, 0); //precision error for compressed, but performance the same
                    }
                    counter++;
                    if ((counter % 100) == 0) {
                        if (useDatabase) {
                            stmt.executeBatch();
                            stmt.clearBatch();
                        }
                        //System.out.println(fileName + " " + limit + " " + counter);
                    }
                }
                stopTime = System.currentTimeMillis(); // stop timing
                in.close();
                time_stats[i].update((stopTime - startTime) / 1000f);

                System.out.println(j + " " + limit + " " + (stopTime - startTime));
            }
            }

            BufferedWriter out_time = new BufferedWriter(new FileWriter(OUTPUT_PATH + "test_update_" + (useDatabase ? "dbms" : "tree") + ".dat"));
            out_time.write("page250\ttime\ttime_stdev"); out_time.newLine();
            for (int i = 0; i < n; i++) {
                out_time.write(String.format(Locale.US, "%d\t%.3f\t%.3f", (i + 1) * 4, time_stats[i].getAvg(), time_stats[i].getStdev()));
                out_time.newLine();
            }
            out_time.close();

        } catch (IOException e) {
            e.printStackTrace();
        }
        catch (SQLException ex) {
            System.out.println("SQLException: " + ex.getMessage());
            System.out.println("SQLState: " + ex.getSQLState());
            System.out.println("VendorError: " + ex.getErrorCode());
            System.out.println("Query: " + query);
        }

    }

    public void chartProgress(ChartProgressEvent event) {
        if (event.getType() == ChartProgressEvent.DRAWING_STARTED) {
            chartLatchTest = new CountDownLatch(1);
        } else if (event.getType() == ChartProgressEvent.DRAWING_FINISHED) {
            chartLatchTest.countDown();
            panelTest.invalidate();
        }
    }

    public void waitForChart() {
        try {
            if (chartLatchTest != null) {
                chartLatchTest.await();
            }
            Thread.sleep(200);
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
    }
    
    public static void gc() {
        Object obj = new Object();
        WeakReference ref = new WeakReference<Object>(obj);
        obj = null;
        while(ref.get() != null) {
            System.gc();
        }
    }

    // Implements WindowListener
    public void windowOpened(WindowEvent windowEvent) { }
    public void windowClosing(WindowEvent e) { finalize(); }
    public void windowClosed(WindowEvent e) { }
    public void windowIconified(WindowEvent e) { }
    public void windowDeiconified(WindowEvent e) { }
    public void windowActivated(WindowEvent e) { repaint(); }
    public void windowDeactivated(WindowEvent e) { }
}
